"""
QCPortal Database ODM
"""
import itertools as it
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union

import numpy as np
import pandas as pd
from qcelemental import constants

from ..models import Molecule, ProtoModel
from ..util import replace_dict_keys
from .collection_utils import nCr, register_collection
from .dataset import Dataset

if TYPE_CHECKING:  # pragma: no cover
    from .. import FractalClient
    from ..models import ComputeResponse


class _ReactionTypeEnum(str, Enum):
    """Helper class for locking the reaction type into one or the other"""

    rxn = "rxn"
    ie = "ie"


class ReactionEntry(ProtoModel):
    """Data model for the `reactions` list in Dataset"""

    attributes: Dict[str, Union[int, float, str]]  # Might be overloaded key types
    reaction_results: Dict[str, dict]
    name: str
    stoichiometry: Dict[str, Dict[str, float]]
    extras: Dict[str, Any] = {}


class ReactionDataset(Dataset):
    """
    The ReactionDataset class for homogeneous computations on many reactions.

    Attributes
    ----------
    client : client.FractalClient
        A FractalClient connected to a server
    data : ReactionDataset.DataModel
        A Model representation of the database backbone
    df : pd.DataFrame
        The underlying dataframe for the Dataset object
    rxn_index : pd.Index
        The unrolled reaction index for all reactions in the Dataset
    """

    def __init__(self, name: str, client: Optional["FractalClient"] = None, ds_type: str = "rxn", **kwargs) -> None:
        """
        Initializer for the Dataset object. If no Portal is supplied or the database name
        is not present on the server that the Portal is connected to a blank database will be
        created.

        Parameters
        ----------
        name : str
            The name of the Dataset
        client : client.FractalClient, optional
            A FractalClient connected to a server
        ds_type : str, optional
            The type of Dataset involved

        """
        ds_type = ds_type.lower()
        super().__init__(name, client=client, ds_type=ds_type, **kwargs)

    class DataModel(Dataset.DataModel):

        ds_type: _ReactionTypeEnum = _ReactionTypeEnum.rxn
        records: Optional[List[ReactionEntry]] = None

        history: Set[Tuple[str, str, str, Optional[str], Optional[str], str]] = set()
        history_keys: Tuple[str, str, str, str, str, str] = (
            "driver",
            "program",
            "method",
            "basis",
            "keywords",
            "stoichiometry",
        )

    def _entry_index(self, subset: Optional[List[str]] = None) -> None:
        if self.data.records is None:
            self._get_data_records_from_db()
        # Unroll the index
        tmp_index = []
        for rxn in self.data.records:
            name = rxn.name
            for stoich_name in list(rxn.stoichiometry):
                for mol_hash, coef in rxn.stoichiometry[stoich_name].items():
                    tmp_index.append([name, stoich_name, mol_hash, coef])
        ret = pd.DataFrame(tmp_index, columns=["name", "stoichiometry", "molecule", "coefficient"])
        if subset is None:
            return ret
        else:
            return ret.reset_index().set_index("name").loc[subset].reset_index().set_index("index")

    def _molecule_indexer(
        self,
        stoich: Union[str, List[str]],
        subset: Optional[Union[str, Set[str]]] = None,
        coefficients: bool = False,
        force: bool = False,
    ) -> Tuple[Dict[Tuple[str, ...], "ObjectId"], Tuple[str]]:
        """Provides a {index: molecule_id} mapping for a given subset.

        Parameters
        ----------
        stoich : Union[str, List[str]]
            The stoichiometries, or list of stoichiometries to return
        subset : Optional[Union[str, Set[str]]], optional
            The indices of the desired subset. Return all indices if subset is None.
        coefficients : bool, optional
            Returns the coefficients if as part of the index if True

        No Longer Returned
        ------------------
        Dict[str, 'ObjectId']
            Molecule index to molecule ObjectId map

        Returns
        -------
        Tuple[Dict[Tuple[str, ...], 'ObjectId'], Tuple[str]]
            Molecule index to molecule ObjectId map, and index names
        """
        if isinstance(stoich, str):
            stoich = [stoich]

        index = self.get_entries(subset=subset, force=force)
        matched_rows = index[np.in1d(index["stoichiometry"], stoich)]

        if subset:
            matched_rows = matched_rows[np.in1d(matched_rows["name"], subset)]

        names = ("name", "stoichiometry", "idx")
        if coefficients:
            names = names + ("coefficient",)

        ret = {}
        for gb_idx, group in matched_rows.groupby(["name", "stoichiometry"]):
            for cnt, (idx, row) in enumerate(group.iterrows()):
                if coefficients:
                    ret[gb_idx + (cnt, row["coefficient"])] = row["molecule"]
                else:
                    ret[gb_idx + (cnt,)] = row["molecule"]

        return ret, names

    def valid_stoich(self, subset=None, force: bool = False) -> Set[str]:
        entries = self.get_entries(subset=subset, force=force)
        return set(entries["stoichiometry"].unique())

    def _validate_stoich(self, stoich: Union[List[str], str], subset=None, force: bool = False) -> None:
        if isinstance(stoich, str):
            stoich = [stoich]
        if isinstance(subset, str):
            subset = [subset]
        valid_stoich = self.valid_stoich(subset=subset, force=force)
        for s in stoich:
            if s.lower() not in valid_stoich:
                raise KeyError("Stoichiometry not understood, valid keys are {}.".format(valid_stoich))

    def _pre_save_prep(self, client: "FractalClient") -> None:
        self._canonical_pre_save(client)

        mol_ret = self._add_molecules_by_dict(client, self._new_molecules)

        # Update internal molecule UUID's to servers UUID's
        for record in self._new_records:
            stoichiometry = replace_dict_keys(record.stoichiometry, mol_ret)
            new_record = record.copy(update={"stoichiometry": stoichiometry})
            self.data.records.append(new_record)

        self._new_records: List[ReactionEntry] = []
        self._new_molecules = {}

        self._entry_index()

    def get_values(
        self,
        method: Optional[Union[str, List[str]]] = None,
        basis: Optional[Union[str, List[str]]] = None,
        keywords: Optional[str] = None,
        program: Optional[str] = None,
        driver: Optional[str] = None,
        stoich: str = "default",
        name: Optional[Union[str, List[str]]] = None,
        native: Optional[bool] = None,
        subset: Optional[Union[str, List[str]]] = None,
        force: bool = False,
    ) -> pd.DataFrame:
        """
        Obtains values from the known history from the search paramaters provided for the expected `return_result` values.
        Defaults to the standard programs and keywords if not provided.

        Note that unlike `get_records`, `get_values` will automatically expand searches and return multiple method
        and basis combinations simultaneously.

        `None` is a wildcard selector. To search for `None`, use `"None"`.

        Parameters
        ----------
        method : Optional[Union[str, List[str]]], optional
            The computational method (B3LYP)
        basis : Optional[Union[str, List[str]]], optional
            The computational basis (6-31G)
        keywords : Optional[str], optional
            The keyword alias
        program : Optional[str], optional
            The underlying QC program
        driver : Optional[str], optional
            The type of calculation (e.g. energy, gradient, hessian, dipole...)
        stoich : str, optional
            Stoichiometry of the reaction.
        name : Optional[Union[str, List[str]]], optional
            Canonical name of the record. Overrides the above selectors.
        native: Optional[bool], optional
            True: only include data computed with QCFractal
            False: only include data contributed from outside sources
            None: include both
        subset: Optional[List[str]], optional
            The indices of the desired subset. Return all indices if subset is None.
        force : bool, optional
            Data is typically cached, forces a new query if True

        Returns
        ------
        DataFrame
           A DataFrame of values with columns corresponding to methods and rows corresponding to reaction entries.
           Contributed (native=False) columns are marked with "(contributed)" and may include units in square brackets
           if their units differ in dimensionality from the ReactionDataset's default units.
        """
        return self._get_values(
            method=method,
            basis=basis,
            keywords=keywords,
            program=program,
            driver=driver,
            stoich=stoich,
            name=name,
            native=native,
            subset=subset,
            force=force,
        )

    def _get_native_values(
        self,
        subset: Set[str],
        method: Optional[str] = None,
        basis: Optional[str] = None,
        keywords: Optional[str] = None,
        program: Optional[str] = None,
        stoich: Optional[str] = None,
        name: Optional[str] = None,
        force: bool = False,
    ) -> pd.DataFrame:
        self._validate_stoich(stoich, subset=subset, force=force)

        # So that datasets with no records do not require a default program and default keywords
        if len(self.list_records()) == 0:
            return pd.DataFrame(index=self.get_index(subset))

        queries = self._form_queries(
            method=method, basis=basis, keywords=keywords, program=program, stoich=stoich, name=name
        )

        if len(queries) == 0:
            return pd.DataFrame(index=self.get_index(subset))

        stoich_complex = queries.pop("stoichiometry").values[0]
        stoich_monomer = "".join([x for x in stoich_complex if not x.isdigit()]) + "1"

        def _query_apply_coeffients(stoich, query):

            # Build the starting table
            indexer, names = self._molecule_indexer(stoich=stoich, coefficients=True, force=force)
            df = self._get_records(indexer, query, include=["return_result"], merge=True)
            df.index = pd.MultiIndex.from_tuples(df.index, names=names)
            df.reset_index(inplace=True)

            # Block out null values `groupby.sum()` will return 0 rather than NaN in all cases
            null_mask = df[["name", "return_result"]].copy()
            null_mask["return_result"] = null_mask["return_result"].isnull()
            null_mask = null_mask.groupby(["name"])["return_result"].sum() != False

            # Multiply by coefficients and sum
            df["return_result"] *= df["coefficient"]
            df = df.groupby(["name"])["return_result"].sum()
            df[null_mask] = np.nan
            return df

        names = []
        new_queries = []
        new_data = pd.DataFrame(index=subset)

        for _, query in queries.iterrows():

            query = query.replace({np.nan: None}).to_dict()
            qname = query["name"]
            names.append(qname)

            if force or not self._subset_in_cache(qname, subset):
                self._column_metadata[qname] = query
                new_queries.append(query)

        if not self._use_view(force):
            units: Dict[str, str] = {}
            for query in new_queries:
                qname = query.pop("name")
                if self.data.ds_type == _ReactionTypeEnum.ie:
                    # This implements 1-body counterpoise correction
                    # TODO: this will need to contain the logic for VMFC or other method-of-increments strategies
                    data_complex = _query_apply_coeffients(stoich_complex, query)
                    data_monomer = _query_apply_coeffients(stoich_monomer, query)
                    data = data_complex - data_monomer
                elif self.data.ds_type == _ReactionTypeEnum.rxn:
                    data = _query_apply_coeffients(stoich_complex, query)
                else:
                    raise ValueError(
                        f"ReactionDataset ds_type is not a member of _ReactionTypeEnum. (Got {self.data.ds_type}.)"
                    )

                new_data[qname] = data * constants.conversion_factor("hartree", self.units)
                query["name"] = qname
                units[qname] = self.units
        else:
            for query in new_queries:
                query["native"] = True
            new_data, units = self._view.get_values(new_queries)
            for query in new_queries:
                qname = query["name"]
                new_data[qname] = new_data[qname] * constants.conversion_factor(units[qname], self.units)

        for query in new_queries:
            qname = query["name"]
            self._column_metadata[qname].update({"native": True, "units": units[qname]})

        self._update_cache(new_data)
        return self.df.loc[subset, names]

    def visualize(
        self,
        method: Optional[str] = None,
        basis: Optional[str] = None,
        keywords: Optional[str] = None,
        program: Optional[str] = None,
        stoich: str = "default",
        groupby: Optional[str] = None,
        metric: str = "UE",
        bench: Optional[str] = None,
        kind: str = "bar",
        return_figure: Optional[bool] = None,
        show_incomplete: bool = False,
    ) -> "plotly.Figure":
        """
        Parameters
        ----------
        method : Optional[str], optional
            Methods to query
        basis : Optional[str], optional
            Bases to query
        keywords : Optional[str], optional
            Keyword aliases to query
        program : Optional[str], optional
            Programs aliases to query
        stoich : str, optional
            Stoichiometry to query
        groupby : Optional[str], optional
            Groups the plot by this index.
        metric : str, optional
            The metric to use either UE (unsigned error) or URE (unsigned relative error)
        bench : Optional[str], optional
            The benchmark level of theory to use
        kind : str, optional
            The kind of chart to produce, either 'bar' or 'violin'
        return_figure : Optional[bool], optional
            If True, return the raw plotly figure. If False, returns a hosted iPlot. If None, return a iPlot display in Jupyter notebook and a raw plotly figure in all other circumstances.
        show_incomplete: bool, optional
            Display statistics method/basis set combinations where results are incomplete

        Returns
        -------
        plotly.Figure
            The requested figure.
        """

        query = {"method": method, "basis": basis, "keywords": keywords, "program": program, "stoichiometry": stoich}
        query = {k: v for k, v in query.items() if v is not None}

        return self._visualize(
            metric,
            bench,
            query=query,
            groupby=groupby,
            return_figure=return_figure,
            kind=kind,
            show_incomplete=show_incomplete,
        )

    def get_molecules(
        self,
        subset: Optional[Union[str, Set[str]]] = None,
        stoich: Union[str, List[str]] = "default",
        force: bool = False,
    ) -> pd.DataFrame:
        """Queries full Molecules from the database.

        Parameters
        ----------
        subset : Optional[Union[str, Set[str]]], optional
            The index subset to query on
        stoich : Union[str, List[str]], optional
            The stoichiometries to pull from, either a single or multiple stoichiometries
        force : bool, optional
            Force pull of molecules from server

        Return
        ------
        pd.DataFrame
            Indexed Molecules which match the stoich and subset string.
        """

        self._check_client()
        self._check_state()
        if isinstance(subset, str):
            subset = [subset]

        self._validate_stoich(stoich, subset=subset, force=force)

        indexer, names = self._molecule_indexer(stoich=stoich, subset=subset, force=force)
        df = self._get_molecules(indexer, force=force)
        df.index = pd.MultiIndex.from_tuples(df.index, names=names)

        return df

    def get_records(
        self,
        method: str,
        basis: Optional[str] = None,
        *,
        keywords: Optional[str] = None,
        program: Optional[str] = None,
        stoich: Union[str, List[str]] = "default",
        include: Optional[List[str]] = None,
        subset: Optional[Union[str, Set[str]]] = None,
    ) -> Union[pd.DataFrame, "ResultRecord"]:
        """
        Queries the local Portal for the requested keys and stoichiometry.

        Parameters
        ----------
        method : str
            The computational method to query on (B3LYP)
        basis : Optional[str], optional
            The computational basis to query on (6-31G)
        keywords : Optional[str], optional
            The option token desired
        program : Optional[str], optional
            The program to query on
        stoich : Union[str, List[str]], optional
            The given stoichiometry to compute.
        include : Optional[Dict[str, bool]], optional
            The attribute project to perform on the query, otherwise returns ResultRecord objects.
        subset : Optional[Union[str, Set[str]]], optional
            The index subset to query on

        Returns
        -------
        Union[pd.DataFrame, 'ResultRecord']
            The name of the queried column

        """

        self._check_client()
        self._check_state()

        method = method.upper()
        if isinstance(stoich, str):
            stoich = [stoich]

        ret = []
        for s in stoich:
            name, _, history = self._default_parameters(program, method, basis, keywords, stoich=s)
            history.pop("stoichiometry")
            indexer, names = self._molecule_indexer(stoich=s, subset=subset, force=True)
            df = self._get_records(
                indexer,
                history,
                include=include,
                merge=False,
                raise_on_plan="`get_records` can only be used for non-composite quantities. You likely queried a DFT+D method or similar that requires a combination of DFT and -D. Please query each piece separately.",
            )
            df = df[0]
            df.index = pd.MultiIndex.from_tuples(df.index, names=names)
            ret.append(df)

        ret = pd.concat(ret)
        ret.sort_index(inplace=True)

        return ret

    def compute(
        self,
        method: str,
        basis: Optional[str] = None,
        *,
        keywords: Optional[str] = None,
        program: Optional[str] = None,
        stoich: str = "default",
        ignore_ds_type: bool = False,
        tag: Optional[str] = None,
        priority: Optional[str] = None,
    ) -> "ComputeResponse":
        """Executes a computational method for all reactions in the Dataset.
        Previously completed computations are not repeated.

        Parameters
        ----------
        method : str
            The computational method to compute (B3LYP)
        basis : Optional[str], optional
            The computational basis to compute (6-31G)
        keywords : Optional[str], optional
            The keyword alias for the requested compute
        program : Optional[str], optional
            The underlying QC program
        stoich : str, optional
            The stoichiometry of the requested compute (cp/nocp/etc)
        ignore_ds_type : bool, optional
            Optionally only compute the "default" geometry
        tag : Optional[str], optional
            The queue tag to use when submitting compute requests.
        priority : Optional[str], optional
            The priority of the jobs low, medium, or high.

        Returns
        -------
        ComputeResponse
            An object that contains the submitted ObjectIds of the new compute. This object has the following fields:
              - ids: The ObjectId's of the task in the order of input molecules
              - submitted: A list of ObjectId's that were submitted to the compute queue
              - existing: A list of ObjectId's of tasks already in the database

        """
        self._check_client()
        self._check_state()

        entry_index = self.get_entries(force=True)

        self._validate_stoich(stoich, subset=None, force=True)
        compute_keys = {"program": program, "method": method, "basis": basis, "keywords": keywords, "stoich": stoich}

        # Figure out molecules that we need
        if (not ignore_ds_type) and (self.data.ds_type.lower() == "ie"):

            monomer_stoich = "".join([x for x in stoich if not x.isdigit()]) + "1"
            tmp_monomer = entry_index[entry_index["stoichiometry"] == monomer_stoich].copy()

            ret1 = self._compute(compute_keys, tmp_monomer["molecule"], tag, priority)

            tmp_complex = entry_index[entry_index["stoichiometry"] == stoich].copy()
            ret2 = self._compute(compute_keys, tmp_complex["molecule"], tag, priority)

            ret = ret1.merge(ret2)
        else:
            tmp_complex = entry_index[entry_index["stoichiometry"] == stoich].copy()
            ret = self._compute(compute_keys, tmp_complex["molecule"], tag, priority)

        # Update the record that this was computed
        self.save()

        return ret

    def get_rxn(self, name: str) -> ReactionEntry:
        """
        Returns the JSON object of a specific reaction.

        Parameters
        ----------
        name : str
            The name of the reaction to query

        Returns
        -------
        ret : dict
            The JSON representation of the reaction

        """

        found = []
        for num, x in enumerate(self.data.records):
            if x.name == name:
                found.append(num)

        if len(found) == 0:
            raise KeyError("Dataset:get_rxn: Reaction name '{}' not found.".format(name))

        if len(found) > 1:
            raise KeyError("Dataset:get_rxn: Multiple reactions of name '{}' found. Dataset failure.".format(name))

        return self.data.records[found[0]]

    # Visualization
    def ternary(self, cvals=None):
        """Plots a ternary diagram of the DataBase if available

        Parameters
        ----------
        cvals : None, optional
            Description

        """
        raise Exception("MPL not avail")

    #        return visualization.Ternary2D(self.df, cvals=cvals)

    # Adders

    def parse_stoichiometry(self, stoichiometry: List[Tuple[Union[Molecule, str], float]]) -> Dict[str, float]:
        """
        Parses a stiochiometry list.

        Parameters
        ----------
        stoichiometry : list
            A list of tuples describing the stoichiometry.

        Returns
        -------
        Dict[str, float]
            A dictionary describing the stoichiometry for use in the database.
            Keys are molecule hashes. Values are stoichiometric coefficients.

        Notes
        -----
        This function attempts to convert the molecule into its corresponding hash. The following will happen depending on the form of the Molecule.
            - Molecule hash - Used directly in the stoichiometry.
            - Molecule class - Hash is obtained and the molecule will be added to the database upon saving.
            - Molecule string - Molecule will be converted to a Molecule class and the same process as the above will occur.


        """

        mol_hashes = []
        mol_values = []

        for line in stoichiometry:
            if len(line) != 2:
                raise KeyError("Dataset: Parse stoichiometry: passed in as a list, must be of key : value type")

            # Get the values
            try:
                mol_values.append(float(line[1]))
            except:
                raise TypeError("Dataset: Parse stoichiometry: must be able to cast second value to a float.")

            # What kind of molecule is it?
            mol = line[0]

            if isinstance(mol, str) and (len(mol) == 40):
                molecule_hash = mol

            elif isinstance(mol, str):
                qcf_mol = Molecule.from_data(mol)

                molecule_hash = qcf_mol.get_hash()

                if molecule_hash not in list(self._new_molecules):
                    self._new_molecules[molecule_hash] = qcf_mol

            elif isinstance(mol, Molecule):
                molecule_hash = mol.get_hash()

                if molecule_hash not in list(self._new_molecules):
                    self._new_molecules[molecule_hash] = mol

            else:
                raise TypeError(
                    "Dataset: Parse stoichiometry: first value must either be a molecule hash, "
                    "a molecule str, or a Molecule class."
                )

            mol_hashes.append(molecule_hash)

        # Sum together the coefficients of duplicates
        ret: Dict[str, float] = {}
        for mol, coef in zip(mol_hashes, mol_values):
            if mol in list(ret):
                ret[mol] += coef
            else:
                ret[mol] = coef

        return ret

    def add_rxn(
        self,
        name: str,
        stoichiometry: Dict[str, List[Tuple[Molecule, float]]],
        reaction_results: Optional[Dict[str, str]] = None,
        attributes: Optional[Dict[str, Union[int, float, str]]] = None,
        other_fields: Optional[Dict[str, Any]] = None,
    ) -> ReactionEntry:
        """
        Adds a reaction to a database object.

        Parameters
        ----------
        name : str
            Name of the reaction.
        stoichiometry : list or dict
            Either a list or dictionary of lists
        reaction_results :  dict or None, Optional, Default: None
            A dictionary of the computed total interaction energy results
        attributes :  dict or None, Optional, Default: None
            A dictionary of attributes to assign to the reaction
        other_fields : dict or None, Optional, Default: None
            A dictionary of additional user defined fields to add to the reaction entry

        Returns
        -------
        ReactionEntry
            A complete specification of the reaction
        """
        if reaction_results is None:
            reaction_results = {}
        if attributes is None:
            attributes = {}
        if other_fields is None:
            other_fields = {}
        rxn_dict: Dict[str, Any] = {"name": name}

        # Set name
        if name in self.get_index():
            raise KeyError(
                "Dataset: Name '{}' already exists. "
                "Please either delete this entry or call the update function.".format(name)
            )

        # Set stoich
        if isinstance(stoichiometry, dict):
            rxn_dict["stoichiometry"] = {}

            if "default" not in list(stoichiometry):
                raise KeyError("Dataset:add_rxn: Stoichiometry dict must have a 'default' key.")

            for k, v in stoichiometry.items():
                rxn_dict["stoichiometry"][k] = self.parse_stoichiometry(v)

        elif isinstance(stoichiometry, (tuple, list)):
            rxn_dict["stoichiometry"] = {}
            rxn_dict["stoichiometry"]["default"] = self.parse_stoichiometry(stoichiometry)
        else:
            raise TypeError("Dataset:add_rxn: Type of stoichiometry input was not recognized:", type(stoichiometry))

        # Set attributes
        if not isinstance(attributes, dict):
            raise TypeError("Dataset:add_rxn: attributes must be a dictionary, not '{}'".format(type(attributes)))

        rxn_dict["attributes"] = attributes

        if not isinstance(other_fields, dict):
            raise TypeError("Dataset:add_rxn: other_fields must be a dictionary, not '{}'".format(type(attributes)))

        rxn_dict["extras"] = other_fields

        if "default" in list(reaction_results):
            rxn_dict["reaction_results"] = reaction_results
        elif isinstance(reaction_results, dict):
            rxn_dict["reaction_results"] = {}
            rxn_dict["reaction_results"]["default"] = reaction_results
        else:
            raise TypeError("Passed in reaction_results not understood.")

        rxn = ReactionEntry(**rxn_dict)
        self._new_records.append(rxn)

        return rxn

    def add_ie_rxn(self, name: str, mol: Molecule, **kwargs) -> ReactionEntry:
        """Add a interaction energy reaction entry to the database. Automatically
        builds CP and no-CP reactions for the fragmented molecule.

        Parameters
        ----------
        name : str
            The name of the reaction
        mol : Molecule
            A molecule with multiple fragments
        **kwargs
            Additional kwargs to pass into `build_id_fragments`.

        Returns
        -------
        ReactionEntry
            A representation of the new reaction.
        """
        reaction_results = kwargs.pop("reaction_results", {})
        attributes = kwargs.pop("attributes", {})
        other_fields = kwargs.pop("other_fields", {})

        stoichiometry = self.build_ie_fragments(mol, name=name, **kwargs)
        return self.add_rxn(
            name, stoichiometry, reaction_results=reaction_results, attributes=attributes, other_fields=other_fields
        )

    @staticmethod
    def build_ie_fragments(mol: Molecule, **kwargs) -> Dict[str, List[Tuple[Molecule, float]]]:
        """
        Build the stoichiometry for an Interaction Energy.

        Parameters
        ----------
        mol : Molecule class or str
            Molecule to fragment.
        do_default : bool
            Create the default (noCP) stoichiometry.
        do_cp : bool
            Create the counterpoise (CP) corrected stoichiometry.
        do_vmfc : bool
            Create the Valiron-Mayer Function Counterpoise (VMFC) corrected stoichiometry.
        max_nbody : int
            The maximum fragment level built, if zero defaults to the maximum number of fragments.

        Notes
        -----

        Returns
        -------
        ret : dict
            A JSON representation of the fragmented molecule.

        """

        do_default = kwargs.pop("do_default", True)
        do_cp = kwargs.pop("do_cp", True)
        do_vmfc = kwargs.pop("do_vmfc", False)
        max_nbody = kwargs.pop("max_nbody", 0)

        if not isinstance(mol, Molecule):

            mol = Molecule.from_data(mol, **kwargs)

        ret = {}

        max_frag = len(mol.fragments)
        if max_nbody == 0:
            max_nbody = max_frag
        if max_frag < 2:
            raise AttributeError("Dataset:build_ie_fragments: Molecule must have at least two fragments.")

        # Build some info
        fragment_range = list(range(max_frag))

        # Loop over the bodis
        for nbody in range(1, max_nbody):
            nocp_tmp = []
            cp_tmp = []
            for k in range(1, nbody + 1):
                take_nk = nCr(max_frag - k - 1, nbody - k)
                sign = (-1) ** (nbody - k)
                coef = take_nk * sign
                for frag in it.combinations(fragment_range, k):
                    if do_default:
                        nocp_tmp.append((mol.get_fragment(frag, orient=True, group_fragments=True), coef))
                    if do_cp:
                        ghost = list(set(fragment_range) - set(frag))
                        cp_tmp.append((mol.get_fragment(frag, ghost, orient=True, group_fragments=True), coef))

            if do_default:
                ret["default" + str(nbody)] = nocp_tmp

            if do_cp:
                ret["cp" + str(nbody)] = cp_tmp

        # VMFC is a special beast
        if do_vmfc:
            raise KeyError("VMFC isnt quite ready for primetime!")

            # ret.update({"vmfc" + str(nbody): [] for nbody in range(1, max_nbody)})
            # nbody_range = list(range(1, max_nbody))
            # for nbody in nbody_range:
            #     for cp_combos in it.combinations(fragment_range, nbody):
            #         basis_tuple = tuple(cp_combos)
            #         for interior_nbody in nbody_range:
            #             for x in it.combinations(cp_combos, interior_nbody):
            #                 ghost = list(set(basis_tuple) - set(x))
            #                 ret["vmfc" + str(interior_nbody)].append((mol.get_fragment(x, ghost), 1.0))

        # Add in the maximal position
        if do_default:
            ret["default"] = [(mol, 1.0)]

        if do_cp:
            ret["cp"] = [(mol, 1.0)]

        # if do_vmfc:
        #     ret["vmfc"] = [(mol, 1.0)]

        return ret


register_collection(ReactionDataset)



"""Add contributed values table

Revision ID: 159ba85908fd
Revises: d5d88ac1d291
Create Date: 2019-11-01 15:39:50.970246

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.dialects import postgresql
from qcfractal.storage_sockets.models.sql_base import MsgpackExt
from qcfractal.storage_sockets.models.collections_models import ContributedValuesORM

import numpy as np

# revision identifiers, used by Alembic.
revision = "159ba85908fd"
down_revision = "d5d88ac1d291"
branch_labels = None
depends_on = None


def migrate_contributed_values_data():

    bind = op.get_bind()
    session = orm.Session(bind=bind)

    # Dataset and reaction datasets tables
    ds_ids_data = session.execute("select id, contributed_values_data from dataset;").fetchall()
    print(f"Migrating datasets with ids: {[ds[0] for ds in ds_ids_data]}")

    rds_ids_data = session.execute("select id, contributed_values_data from reaction_dataset;").fetchall()
    print(f"Migrating reaction datasets with ids: {[ds[0] for ds in rds_ids_data]}")

    ds_ids_data.extend(rds_ids_data)

    for ds in ds_ids_data:
        (ds_id, ds_contrib) = ds
        if ds_contrib is None:
            continue

        for key, dict_values in ds_contrib.items():

            idx, vals = [], []
            for key, value in dict_values["values"].items():
                idx.append(key)
                vals.append(value)

            dict_values["values"] = np.array(vals)
            dict_values["index"] = np.array(idx)

            cv = ContributedValuesORM(**dict_values)
            cv.collection_id = ds_id

            session.add(cv)

    session.commit()


def upgrade():

    # rename old column with data
    op.alter_column("dataset", "contributed_values", new_column_name="contributed_values_data")
    op.alter_column("reaction_dataset", "contributed_values", new_column_name="contributed_values_data")

    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "contributed_values",
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("collection_id", sa.Integer(), nullable=False),
        sa.Column("citations", sa.JSON(), nullable=True),
        sa.Column("theory_level", sa.JSON(), nullable=False),
        sa.Column("theory_level_details", sa.JSON(), nullable=True),
        sa.Column("comments", sa.String(), nullable=True),
        sa.Column("values", MsgpackExt(), nullable=False),
        sa.Column("index", MsgpackExt(), nullable=False),
        sa.Column("external_url", sa.String(), nullable=True),
        sa.Column("doi", sa.String(), nullable=True),
        sa.Column("units", sa.String(), nullable=False),
        sa.Column("values_structure", sa.JSON(), nullable=True, default=lambda: {}),
        sa.ForeignKeyConstraint(["collection_id"], ["collection.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("name", "collection_id"),
    )

    op.alter_column("contributed_values", "values_structure", server_default=None, nullable=False)

    migrate_contributed_values_data()

    op.drop_column("dataset", "contributed_values_data")
    op.drop_column("reaction_dataset", "contributed_values_data")


def downgrade():
    # ### Won't work on production data because data will be lost ###

    op.add_column(
        "reaction_dataset",
        sa.Column("contributed_values", postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True),
    )
    op.add_column(
        "dataset",
        sa.Column("contributed_values", postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True),
    )
    op.drop_table("contributed_values")



"""Add/fix some indices on base_result and molecule

Revision ID: c05d63683601
Revises: 5be555fe9dc0
Create Date: 2021-03-04 11:29:11.780381

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "c05d63683601"
down_revision = "5be555fe9dc0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index("ix_base_result_error", "base_result", ["error"], unique=True)
    op.create_index("ix_base_result_stderr", "base_result", ["stderr"], unique=True)
    op.create_index("ix_base_result_stdout", "base_result", ["stdout"], unique=True)

    # The molecule table may contain duplicates. Don't do this for now
    # op.drop_index("ix_molecule_hash", table_name="molecule")
    # op.create_index("ix_molecule_hash", "molecule", ["molecule_hash"], unique=True)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # op.drop_index("ix_molecule_hash", table_name="molecule")
    # op.create_index("ix_molecule_hash", "molecule", ["molecule_hash"], unique=False)
    op.drop_index("ix_base_result_stdout", table_name="base_result")
    op.drop_index("ix_base_result_stderr", table_name="base_result")
    op.drop_index("ix_base_result_error", table_name="base_result")
    # ### end Alembic commands ###



"""Add indices to access and base_result

Revision ID: cd3d3e2915ce
Revises: c05d63683601
Create Date: 2021-03-08 09:40:56.987985

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "cd3d3e2915ce"
down_revision = "c05d63683601"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("access_type", table_name="access_log")
    op.create_index(op.f("ix_access_log_access_date"), "access_log", ["access_date"], unique=False)
    op.create_index(op.f("ix_access_log_access_type"), "access_log", ["access_type"], unique=False)
    op.create_index("ix_base_result_hash_index", "base_result", ["hash_index"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_base_result_hash_index", table_name="base_result")
    op.drop_index(op.f("ix_access_log_access_type"), table_name="access_log")
    op.drop_index(op.f("ix_access_log_access_date"), table_name="access_log")
    op.create_index("access_type", "access_log", ["access_date"], unique=False)
    # ### end Alembic commands ###



"""
Helpers to hash complex objects
"""

import numpy as np


def float_prep(array, around):
    """
    Rounds floats to a common value and build positive zero's to prevent hash conflicts.
    """

    if isinstance(array, (list, np.ndarray)):
        # Round array
        array = np.around(array, around)
        # Flip zeros
        array[np.abs(array) < 5 ** (-(around + 1))] = 0

    elif isinstance(array, (float, int)):
        array = round(array, around)
        if array == -0.0:
            array = 0.0
    else:
        raise TypeError("Type '{}' not recognized".format(type(array).__name__))

    return array



import datetime

from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Index,
    Integer,
    String,
    UniqueConstraint,
    func,
    select,
)
from sqlalchemy.dialects.postgresql import JSONB, aggregate_order_by
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import column_property, relationship

from qcfractal.interface.models.records import DriverEnum, RecordStatusEnum
from qcfractal.storage_sockets.models.sql_base import Base, MsgpackExt
from qcfractal.storage_sockets.models.sql_models import KeywordsORM, KVStoreORM, MoleculeORM


class BaseResultORM(Base):
    """
    Abstract Base class for ResultORMs and ProcedureORMs
    """

    __tablename__ = "base_result"

    # for SQL
    result_type = Column(String)  # for inheritance

    # Base identification
    id = Column(Integer, primary_key=True)
    # ondelete="SET NULL": when manger is deleted, set this field to None
    manager_name = Column(String, ForeignKey("queue_manager.name", ondelete="SET NULL"), nullable=True)

    hash_index = Column(String)  # TODO
    procedure = Column(String(100), nullable=False)  # TODO: may remove
    version = Column(Integer)
    protocols = Column(JSONB, nullable=False)

    # Extra fields
    extras = Column(MsgpackExt)
    stdout = Column(Integer, ForeignKey("kv_store.id"))
    stdout_obj = relationship(
        KVStoreORM, lazy="noload", foreign_keys=stdout, cascade="all, delete-orphan", single_parent=True
    )

    stderr = Column(Integer, ForeignKey("kv_store.id"))
    stderr_obj = relationship(
        KVStoreORM, lazy="noload", foreign_keys=stderr, cascade="all, delete-orphan", single_parent=True
    )

    error = Column(Integer, ForeignKey("kv_store.id"))
    error_obj = relationship(
        KVStoreORM, lazy="noload", foreign_keys=error, cascade="all, delete-orphan", single_parent=True
    )

    # Compute status
    status = Column(Enum(RecordStatusEnum), nullable=False, default=RecordStatusEnum.incomplete)

    created_on = Column(DateTime, default=datetime.datetime.utcnow)
    modified_on = Column(DateTime, default=datetime.datetime.utcnow)

    # Carry-ons
    provenance = Column(JSON)

    __table_args__ = (
        Index("ix_base_result_status", "status"),
        Index("ix_base_result_type", "result_type"),  # todo: needed?
        Index("ix_base_result_stdout", "stdout", unique=True),
        Index("ix_base_result_stderr", "stderr", unique=True),
        Index("ix_base_result_error", "error", unique=True),
        Index("ix_base_result_hash_index", "hash_index", unique=False),
    )

    __mapper_args__ = {"polymorphic_on": "result_type"}


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class WavefunctionStoreORM(Base):

    __tablename__ = "wavefunction_store"

    id = Column(Integer, primary_key=True)

    # Sparsity is very cheap
    basis = Column(MsgpackExt, nullable=False)
    restricted = Column(Boolean, nullable=False)

    # Core Hamiltonian
    h_core_a = Column(MsgpackExt, nullable=True)
    h_core_b = Column(MsgpackExt, nullable=True)
    h_effective_a = Column(MsgpackExt, nullable=True)
    h_effective_b = Column(MsgpackExt, nullable=True)

    # SCF Results
    scf_orbitals_a = Column(MsgpackExt, nullable=True)
    scf_orbitals_b = Column(MsgpackExt, nullable=True)
    scf_density_a = Column(MsgpackExt, nullable=True)
    scf_density_b = Column(MsgpackExt, nullable=True)
    scf_fock_a = Column(MsgpackExt, nullable=True)
    scf_fock_b = Column(MsgpackExt, nullable=True)
    scf_eigenvalues_a = Column(MsgpackExt, nullable=True)
    scf_eigenvalues_b = Column(MsgpackExt, nullable=True)
    scf_occupations_a = Column(MsgpackExt, nullable=True)
    scf_occupations_b = Column(MsgpackExt, nullable=True)

    # Extras
    extras = Column(MsgpackExt, nullable=True)


class ResultORM(BaseResultORM):
    """
    Hold the result of an atomic single calculation
    """

    __tablename__ = "result"

    id = Column(Integer, ForeignKey("base_result.id", ondelete="CASCADE"), primary_key=True)

    # uniquely identifying a result
    program = Column(String(100), nullable=False)  # example "rdkit", is it the same as program in keywords?
    driver = Column(String(100), Enum(DriverEnum), nullable=False)
    method = Column(String(100), nullable=False)  # example "uff"
    basis = Column(String(100))
    molecule = Column(Integer, ForeignKey("molecule.id"), nullable=False)
    molecule_obj = relationship(MoleculeORM, lazy="select")

    # This is a special case where KeywordsORM are denormalized intentionally as they are part of the
    # lookup for a single result and querying a result will not often request the keywords (LazyReference)
    keywords = Column(Integer, ForeignKey("keywords.id"))
    keywords_obj = relationship(KeywordsORM, lazy="select")

    # Primary Result output
    return_result = Column(MsgpackExt)
    properties = Column(JSON)  # TODO: may use JSONB in the future

    # Wavefunction data
    wavefunction = Column(JSONB, nullable=True)
    wavefunction_data_id = Column(Integer, ForeignKey("wavefunction_store.id"), nullable=True)
    wavefunction_data_obj = relationship(
        WavefunctionStoreORM,
        lazy="noload",
        foreign_keys=wavefunction_data_id,
        cascade="all, delete-orphan",
        single_parent=True,
    )

    __table_args__ = (
        # TODO: optimize indexes
        # A multicolumn GIN index can be used with query conditions that involve any subset of
        # the index's columns. Unlike B-tree or GiST, index search effectiveness is the same
        # regardless of which index column(s) the query conditions use.
        # Index('ix_result_combined', "program", "driver", "method", "basis",
        #       "keywords", postgresql_using='gin'),  # gin index
        # Index('ix_results_molecule', 'molecule'),  # b-tree index
        UniqueConstraint("program", "driver", "method", "basis", "keywords", "molecule", name="uix_results_keys"),
    )

    __mapper_args__ = {
        "polymorphic_identity": "result",
        # to have separate select when querying BaseResultsORM
        "polymorphic_load": "selectin",
    }


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class ProcedureMixin:
    """
    A procedure mixin to be used by specific procedure types
    """

    program = Column(String(100), nullable=False)
    keywords = Column(JSON)
    qc_spec = Column(JSON)


# ================== Types of ProcedureORMs ================== #


class Trajectory(Base):
    """Association table for many to many"""

    __tablename__ = "opt_result_association"

    opt_id = Column(Integer, ForeignKey("optimization_procedure.id", ondelete="cascade"), primary_key=True)
    result_id = Column(Integer, ForeignKey("result.id", ondelete="cascade"), primary_key=True)
    position = Column(Integer, primary_key=True)
    # Index('opt_id', 'result_id', unique=True)

    # trajectory_obj = relationship(ResultORM, lazy="noload")


class OptimizationProcedureORM(ProcedureMixin, BaseResultORM):
    """
    An Optimization  procedure
    """

    __tablename__ = "optimization_procedure"

    id = Column(Integer, ForeignKey("base_result.id", ondelete="cascade"), primary_key=True)

    def __init__(self, **kwargs):
        kwargs.setdefault("version", 1)
        self.procedure = "optimization"
        super().__init__(**kwargs)

    schema_version = Column(Integer, default=1)
    initial_molecule = Column(Integer, ForeignKey("molecule.id"), nullable=False)
    initial_molecule_obj = relationship(MoleculeORM, lazy="select", foreign_keys=initial_molecule)

    # # Results
    energies = Column(JSON)  # Column(ARRAY(Float))
    final_molecule = Column(Integer, ForeignKey("molecule.id"))
    final_molecule_obj = relationship(MoleculeORM, lazy="select", foreign_keys=final_molecule)

    # ids, calculated not stored in this table
    # NOTE: this won't work in SQLite since it returns ARRAYS, aggregate_order_by
    trajectory = column_property(
        select([func.array_agg(aggregate_order_by(Trajectory.result_id, Trajectory.position))]).where(
            Trajectory.opt_id == id
        )
    )

    # array of objects (results) - Lazy - raise error of accessed
    trajectory_obj = relationship(
        Trajectory,
        cascade="all, delete-orphan",
        # backref="optimization_procedure",
        order_by=Trajectory.position,
        collection_class=ordering_list("position"),
    )

    __mapper_args__ = {
        "polymorphic_identity": "optimization_procedure",
        # to have separate select when querying BaseResultsORM
        "polymorphic_load": "selectin",
    }

    __table_args__ = (Index("ix_optimization_program", "program"),)  # todo: needed for procedures?

    def update_relations(self, trajectory=None, **kwarg):

        # update optimization_results relations
        # self._update_many_to_many(opt_result_association, 'opt_id', 'result_id',
        #                 self.id, trajectory, self.trajectory)

        self.trajectory_obj = []
        trajectory = [] if not trajectory else trajectory
        for result_id in trajectory:
            traj = Trajectory(opt_id=int(self.id), result_id=int(result_id))
            self.trajectory_obj.append(traj)

    # def add_relations(self, trajectory):
    #     session = object_session(self)
    #     # add many to many relation with results if ids are given not objects
    #     if trajectory:
    #         session.execute(
    #             opt_result_association
    #                 .insert()  # or update
    #                 .values([(self.id, i) for i in trajectory])
    #         )
    #     session.commit()


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class GridOptimizationAssociation(Base):
    """Association table for many to many"""

    __tablename__ = "grid_optimization_association"

    grid_opt_id = Column(Integer, ForeignKey("grid_optimization_procedure.id", ondelete="cascade"), primary_key=True)
    key = Column(String, nullable=False, primary_key=True)

    # not primary key
    opt_id = Column(Integer, ForeignKey("optimization_procedure.id", ondelete="cascade"))

    # Index('grid_opt_id', 'key', unique=True)

    # optimization_obj = relationship(OptimizationProcedureORM, lazy="joined")


class GridOptimizationProcedureORM(ProcedureMixin, BaseResultORM):

    __tablename__ = "grid_optimization_procedure"

    id = Column(Integer, ForeignKey("base_result.id", ondelete="cascade"), primary_key=True)

    def __init__(self, **kwargs):
        kwargs.setdefault("version", 1)
        kwargs.setdefault("procedure", "gridoptimization")
        kwargs.setdefault("program", "qcfractal")
        super().__init__(**kwargs)

    # Input data
    initial_molecule = Column(Integer, ForeignKey("molecule.id"), nullable=False)
    initial_molecule_obj = relationship(MoleculeORM, lazy="select", foreign_keys=initial_molecule)

    optimization_spec = Column(JSON)

    # Output data
    starting_molecule = Column(Integer, ForeignKey("molecule.id"))
    starting_molecule_obj = relationship(MoleculeORM, lazy="select", foreign_keys=initial_molecule)

    final_energy_dict = Column(JSON)  # Dict[str, float]
    starting_grid = Column(JSON)  # tuple

    grid_optimizations_obj = relationship(
        GridOptimizationAssociation,
        lazy="selectin",
        cascade="all, delete-orphan",
        backref="grid_optimization_procedure",
    )

    @hybrid_property
    def grid_optimizations(self):
        """calculated property when accessed, not saved in the DB
        A view of the many to many relation in the form of a dict"""

        return self._grid_optimizations(self.grid_optimizations_obj)

    @staticmethod
    def _grid_optimizations(grid_optimizations_obj):

        if not grid_optimizations_obj:
            return {}

        if not isinstance(grid_optimizations_obj, list):
            grid_optimizations_obj = [grid_optimizations_obj]

        ret = {}
        try:
            for obj in grid_optimizations_obj:
                ret[obj.key] = str(obj.opt_id)

        except Exception as err:
            # raises exception of first access!!
            pass
            # print(err)

        return ret

    @grid_optimizations.setter
    def grid_optimizations(self, dict_values):

        return dict_values

    __table_args__ = (Index("ix_grid_optmization_program", "program"),)  # todo: needed for procedures?

    __mapper_args__ = {
        "polymorphic_identity": "grid_optimization_procedure",
        # to have separate select when querying BaseResultsORM
        "polymorphic_load": "selectin",
    }

    def update_relations(self, grid_optimizations=None, **kwarg):

        self.grid_optimizations_obj = []
        for key, opt_id in grid_optimizations.items():
            obj = GridOptimizationAssociation(grid_opt_id=int(self.id), opt_id=int(opt_id), key=key)
            self.grid_optimizations_obj.append(obj)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class OptimizationHistory(Base):
    """Association table for many to many"""

    __tablename__ = "optimization_history"

    torsion_id = Column(Integer, ForeignKey("torsiondrive_procedure.id", ondelete="cascade"), primary_key=True)
    opt_id = Column(Integer, ForeignKey("optimization_procedure.id", ondelete="cascade"), primary_key=True)
    key = Column(String, nullable=False, primary_key=True)
    position = Column(Integer, primary_key=True)
    # Index('torsion_id', 'key', unique=True)

    # optimization_obj = relationship(OptimizationProcedureORM, lazy="joined")


class TorsionInitMol(Base):
    """
    Association table for many to many relation
    """

    __tablename__ = "torsion_init_mol_association"

    torsion_id = Column(
        "torsion_id", Integer, ForeignKey("torsiondrive_procedure.id", ondelete="cascade"), primary_key=True
    )
    molecule_id = Column("molecule_id", Integer, ForeignKey("molecule.id", ondelete="cascade"), primary_key=True)


class TorsionDriveProcedureORM(ProcedureMixin, BaseResultORM):
    """
    A torsion drive  procedure
    """

    __tablename__ = "torsiondrive_procedure"

    id = Column(Integer, ForeignKey("base_result.id", ondelete="cascade"), primary_key=True)

    def __init__(self, **kwargs):
        kwargs.setdefault("version", 1)
        self.procedure = "torsiondrive"
        self.program = "torsiondrive"
        super().__init__(**kwargs)

    # input data (along with the mixin)

    # ids of the many to many relation
    initial_molecule = column_property(
        select([func.array_agg(TorsionInitMol.molecule_id)]).where(TorsionInitMol.torsion_id == id)
    )
    # actual objects relation M2M, never loaded here
    initial_molecule_obj = relationship(MoleculeORM, secondary=TorsionInitMol.__table__, uselist=True, lazy="noload")

    optimization_spec = Column(JSON)

    # Output data
    final_energy_dict = Column(JSON)
    minimum_positions = Column(JSON)

    optimization_history_obj = relationship(
        OptimizationHistory,
        cascade="all, delete-orphan",  # backref="torsiondrive_procedure",
        order_by=OptimizationHistory.position,
        collection_class=ordering_list("position"),
        lazy="selectin",
    )

    @hybrid_property
    def optimization_history(self):
        """calculated property when accessed, not saved in the DB
        A view of the many to many relation in the form of a dict"""

        return self._optimization_history(self.optimization_history_obj)

    @staticmethod
    def _optimization_history(optimization_history_obj):

        if not optimization_history_obj:
            return {}

        if not isinstance(optimization_history_obj, list):
            optimization_history_obj = [optimization_history_obj]

        ret = {}
        try:
            for opt_history in optimization_history_obj:
                if opt_history.key in ret:
                    ret[opt_history.key].append(str(opt_history.opt_id))
                else:
                    ret[opt_history.key] = [str(opt_history.opt_id)]

        except Exception as err:
            # raises exception of first access!!
            pass
            # print(err)

        return ret

    @optimization_history.setter
    def optimization_history(self, dict_values):
        """A private copy of the opt history as a dict
        Key: list of optimization procedures"""

        return dict_values

    __table_args__ = (Index("ix_torsion_drive_program", "program"),)  # todo: needed for procedures?

    __mapper_args__ = {
        "polymorphic_identity": "torsiondrive_procedure",
        # to have separate select when querying BaseResultsORM
        "polymorphic_load": "selectin",
    }

    def update_relations(self, initial_molecule=None, optimization_history=None, **kwarg):

        # update torsion molecule relation
        self._update_many_to_many(
            TorsionInitMol.__table__, "torsion_id", "molecule_id", self.id, initial_molecule, self.initial_molecule
        )

        self.optimization_history_obj = []
        for key in optimization_history:
            for opt_id in optimization_history[key]:
                opt_history = OptimizationHistory(torsion_id=int(self.id), opt_id=int(opt_id), key=key)
                self.optimization_history_obj.append(opt_history)

        # No need for the following because the session is committed with parent save
        # session.add_all(self.optimization_history_obj)
        # session.add(self)
        # session.commit()


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



"""
Contains testing infrastructure for QCFractal.
"""

import os
import pkgutil
import shutil
import signal
import socket
import subprocess
import sys
import threading
import time
from collections import Mapping
from contextlib import contextmanager

import numpy as np
import pandas as pd
import pytest
import qcengine as qcng
import requests
from qcelemental.models import Molecule
from tornado.ioloop import IOLoop

from .interface import FractalClient
from .postgres_harness import PostgresHarness, TemporaryPostgres
from .queue import build_queue_adapter
from .server import FractalServer
from .snowflake import FractalSnowflake
from .storage_sockets import storage_socket_factory

### Addon testing capabilities


def pytest_addoption(parser):
    """
    Additional PyTest CLI flags to add

    See `pytest_collection_modifyitems` for handling and `pytest_configure` for adding known in-line marks.
    """
    parser.addoption("--runslow", action="store_true", default=False, help="run slow tests")
    parser.addoption("--runexamples", action="store_true", default=False, help="run example tests")


def pytest_collection_modifyitems(config, items):
    """
    Handle test triggers based on the CLI flags

    Use decorators:
    @pytest.mark.slow
    @pyrest.mark.example
    """
    runslow = config.getoption("--runslow")
    runexamples = config.getoption("--runexamples")
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    skip_example = pytest.mark.skip(reason="need --runexamples option to run")
    for item in items:
        if "slow" in item.keywords and not runslow:
            item.add_marker(skip_slow)
        if "example" in item.keywords and not runexamples:
            item.add_marker(skip_example)


def pytest_configure(config):
    import sys

    sys._called_from_test = True
    config.addinivalue_line("markers", "example: Mark a given test as an example which can be run")
    config.addinivalue_line(
        "markers", "slow: Mark a given test as slower than most other tests, needing a special " "flag to run."
    )


def pytest_unconfigure(config):
    import sys

    del sys._called_from_test


def _plugin_import(plug):
    plug_spec = pkgutil.find_loader(plug)
    if plug_spec is None:
        return False
    else:
        return True


_adapter_testing = ["pool", "dask", "fireworks", "parsl"]

# Figure out what is imported
_programs = {
    "fireworks": _plugin_import("fireworks"),
    "rdkit": _plugin_import("rdkit"),
    "psi4": _plugin_import("psi4"),
    "parsl": _plugin_import("parsl"),
    "dask": _plugin_import("dask"),
    "dask_jobqueue": _plugin_import("dask_jobqueue"),
    "geometric": _plugin_import("geometric"),
    "torsiondrive": _plugin_import("torsiondrive"),
    "torchani": _plugin_import("torchani"),
}
if _programs["dask"]:
    _programs["dask.distributed"] = _plugin_import("dask.distributed")
else:
    _programs["dask.distributed"] = False

_programs["dftd3"] = "dftd3" in qcng.list_available_programs()


def has_module(name):
    return _programs[name]


def check_has_module(program):
    import_message = "Not detecting module {}. Install package if necessary to enable tests."
    if has_module(program) is False:
        pytest.skip(import_message.format(program))


def _build_pytest_skip(program):
    import_message = "Not detecting module {}. Install package if necessary to enable tests."
    return pytest.mark.skipif(has_module(program) is False, reason=import_message.format(program))


# Add a number of module testing options
using_dask = _build_pytest_skip("dask.distributed")
using_dask_jobqueue = _build_pytest_skip("dask_jobqueue")
using_dftd3 = _build_pytest_skip("dftd3")
using_fireworks = _build_pytest_skip("fireworks")
using_geometric = _build_pytest_skip("geometric")
using_parsl = _build_pytest_skip("parsl")
using_psi4 = _build_pytest_skip("psi4")
using_rdkit = _build_pytest_skip("rdkit")
using_torsiondrive = _build_pytest_skip("torsiondrive")
using_unix = pytest.mark.skipif(
    os.name.lower() != "posix", reason="Not on Unix operating system, " "assuming Bash is not present"
)

### Generic helpers


def recursive_dict_merge(base_dict, dict_to_merge_in):
    """Recursive merge for more complex than a simple top-level merge {**x, **y} which does not handle nested dict."""
    for k, v in dict_to_merge_in.items():
        if k in base_dict and isinstance(base_dict[k], dict) and isinstance(dict_to_merge_in[k], Mapping):
            recursive_dict_merge(base_dict[k], dict_to_merge_in[k])
        else:
            base_dict[k] = dict_to_merge_in[k]


def find_open_port():
    """
    Use socket's built in ability to find an open port.
    """
    sock = socket.socket()
    sock.bind(("", 0))

    host, port = sock.getsockname()

    return port


@contextmanager
def preserve_cwd():
    """Always returns to CWD on exit"""
    cwd = os.getcwd()
    try:
        yield cwd
    finally:
        os.chdir(cwd)


def await_true(wait_time, func, *args, **kwargs):

    wait_period = kwargs.pop("period", 4)
    periods = max(int(wait_time / wait_period), 1)
    for period in range(periods):
        ret = func(*args, **kwargs)
        if ret:
            return True
        time.sleep(wait_period)

    return False


### Background thread loops


@contextmanager
def pristine_loop():
    """
    Builds a clean IOLoop for using as a background request.
    Courtesy of Dask Distributed
    """
    IOLoop.clear_instance()
    IOLoop.clear_current()
    loop = IOLoop()
    loop.make_current()
    assert IOLoop.current() is loop

    try:
        yield loop
    finally:
        try:
            loop.close(all_fds=True)
        except (ValueError, KeyError, RuntimeError):
            pass
        IOLoop.clear_instance()
        IOLoop.clear_current()


@contextmanager
def loop_in_thread():
    with pristine_loop() as loop:
        # Add the IOloop to a thread daemon
        thread = threading.Thread(target=loop.start, name="test IOLoop")
        thread.daemon = True
        thread.start()
        loop_started = threading.Event()
        loop.add_callback(loop_started.set)
        loop_started.wait()

        try:
            yield loop
        finally:
            try:
                loop.add_callback(loop.stop)
                thread.join(timeout=5)
            except:
                pass


def terminate_process(proc):
    if proc.poll() is None:

        # Sigint (keyboard interupt)
        if sys.platform.startswith("win"):
            proc.send_signal(signal.CTRL_BREAK_EVENT)
        else:
            proc.send_signal(signal.SIGINT)

        try:
            start = time.time()
            while (proc.poll() is None) and (time.time() < (start + 15)):
                time.sleep(0.02)
        # Flat kill
        finally:
            proc.kill()


@contextmanager
def popen(args, **kwargs):
    """
    Opens a background task.

    Code and idea from dask.distributed's testing suite
    https://github.com/dask/distributed
    """
    args = list(args)

    # Bin prefix
    if sys.platform.startswith("win"):
        bin_prefix = os.path.join(sys.prefix, "Scripts")
    else:
        bin_prefix = os.path.join(sys.prefix, "bin")

    # Do we prefix with Python?
    if kwargs.pop("append_prefix", True):
        args[0] = os.path.join(bin_prefix, args[0])

    # Add coverage testing
    if kwargs.pop("coverage", False):
        coverage_dir = os.path.join(bin_prefix, "coverage")
        if not os.path.exists(coverage_dir):
            print("Could not find Python coverage, skipping cov.")

        else:
            src_dir = os.path.dirname(os.path.abspath(__file__))
            coverage_flags = [coverage_dir, "run", "--parallel-mode", "--source=" + src_dir]

            # If python script, skip the python bin
            if args[0].endswith("python"):
                args.pop(0)
            args = coverage_flags + args

    # Do we optionally dumpstdout?
    dump_stdout = kwargs.pop("dump_stdout", False)

    if sys.platform.startswith("win"):
        # Allow using CTRL_C_EVENT / CTRL_BREAK_EVENT
        kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP

    kwargs["stdout"] = subprocess.PIPE
    kwargs["stderr"] = subprocess.PIPE
    proc = subprocess.Popen(args, **kwargs)
    try:
        yield proc
    except Exception:
        dump_stdout = True
        raise

    finally:
        try:
            terminate_process(proc)
        finally:
            output, error = proc.communicate()
            if dump_stdout:
                print("\n" + "-" * 30)
                print("\n|| Process command: {}".format(" ".join(args)))
                print("\n|| Process stderr: \n{}".format(error.decode()))
                print("-" * 30)
                print("\n|| Process stdout: \n{}".format(output.decode()))
                print("-" * 30)


def run_process(args, **kwargs):
    """
    Runs a process in the background until complete.

    Returns True if exit code zero.
    """

    timeout = kwargs.pop("timeout", 30)
    terminate_after = kwargs.pop("interupt_after", None)
    with popen(args, **kwargs) as proc:
        if terminate_after is None:
            proc.wait(timeout=timeout)
        else:
            time.sleep(terminate_after)
            terminate_process(proc)

        retcode = proc.poll()

    return retcode == 0


### Server testing mechanics


@pytest.fixture(scope="session")
def postgres_server():

    if shutil.which("psql") is None:
        pytest.skip("Postgres is not installed on this server and no active postgres could be found.")

    storage = None
    psql = PostgresHarness({"database": {"port": 5432}})
    # psql = PostgresHarness({"database": {"port": 5432, "username": "qcarchive", "password": "mypass"}})
    if not psql.is_alive():
        print()
        print(
            f"Could not connect to a Postgres server at {psql.config.database_uri()}, this will increase time per test session by ~3 seconds."
        )
        print()
        storage = TemporaryPostgres()
        psql = storage.psql
        print("Using Database: ", psql.config.database_uri())

    yield psql

    if storage:
        storage.stop()


def reset_server_database(server):
    """Resets the server database for testing."""
    if "QCFRACTAL_RESET_TESTING_DB" in os.environ:
        server.storage._clear_db(server.storage._project_name)

    server.storage._delete_DB_data(server.storage._project_name)

    # Force a heartbeat after database clean if a manager is present.
    if server.queue_socket:
        server.await_results()


@pytest.fixture(scope="module")
def test_server(request, postgres_server):
    """
    Builds a server instance with the event loop running in a thread.
    """

    # Storage name
    storage_name = "test_qcfractal_server"
    postgres_server.create_database(storage_name)

    with FractalSnowflake(
        max_workers=0,
        storage_project_name="test_qcfractal_server",
        storage_uri=postgres_server.database_uri(),
        start_server=False,
        reset_database=True,
    ) as server:

        # Clean and re-init the database
        yield server


def build_adapter_clients(mtype, storage_name="test_qcfractal_compute_server"):

    # Basic boot and loop information
    if mtype == "pool":
        from concurrent.futures import ProcessPoolExecutor

        adapter_client = ProcessPoolExecutor(max_workers=2)

    elif mtype == "dask":
        dd = pytest.importorskip("dask.distributed")
        adapter_client = dd.Client(n_workers=2, threads_per_worker=1, resources={"process": 1})

        # Not super happy about this line, but shuts up dangling reference errors
        adapter_client._should_close_loop = False

    elif mtype == "fireworks":
        fireworks = pytest.importorskip("fireworks")

        fireworks_name = storage_name + "_fireworks_queue"
        adapter_client = fireworks.LaunchPad(name=fireworks_name, logdir="/tmp/", strm_lvl="CRITICAL")

    elif mtype == "parsl":
        parsl = pytest.importorskip("parsl")

        # Must only be a single thread as we run thread unsafe applications.
        adapter_client = parsl.config.Config(executors=[parsl.executors.threads.ThreadPoolExecutor(max_threads=1)])

    else:
        raise TypeError("fractal_compute_server: internal parametrize error")

    return adapter_client


@pytest.fixture(scope="module", params=_adapter_testing)
def adapter_client_fixture(request):
    adapter_client = build_adapter_clients(request.param)
    yield adapter_client

    # Do a final close with existing tech
    build_queue_adapter(adapter_client).close()


@pytest.fixture(scope="module", params=_adapter_testing)
def managed_compute_server(request, postgres_server):
    """
    A FractalServer with compute associated parametrize for all managers.
    """

    storage_name = "test_qcfractal_compute_server"
    postgres_server.create_database(storage_name)

    adapter_client = build_adapter_clients(request.param, storage_name=storage_name)

    # Build a server with the thread in a outer context loop
    # Not all adapters play well with internal loops
    with loop_in_thread() as loop:
        server = FractalServer(
            port=find_open_port(),
            storage_project_name=storage_name,
            storage_uri=postgres_server.database_uri(),
            loop=loop,
            queue_socket=adapter_client,
            ssl_options=False,
            skip_storage_version_check=True,
        )

        # Clean and re-init the database
        reset_server_database(server)

        # Build Client and Manager
        from qcfractal.interface import FractalClient

        client = FractalClient(server)

        from qcfractal.queue import QueueManager

        manager = QueueManager(client, adapter_client)

        yield client, server, manager

        # Close down and clean the adapter
        manager.close_adapter()
        manager.stop()


@pytest.fixture(scope="module")
def fractal_compute_server(postgres_server):
    """
    A FractalServer with a local Pool manager.
    """

    # Storage name
    storage_name = "test_qcfractal_compute_snowflake"
    postgres_server.create_database(storage_name)

    with FractalSnowflake(
        max_workers=2,
        storage_project_name=storage_name,
        storage_uri=postgres_server.database_uri(),
        reset_database=True,
        start_server=False,
    ) as server:
        reset_server_database(server)
        yield server


def build_socket_fixture(stype, server=None):
    print("")

    # Check mongo
    storage_name = "test_qcfractal_storage" + stype

    # IP/port/drop table is specific to build
    if stype == "sqlalchemy":

        server.create_database(storage_name)
        storage = storage_socket_factory(server.database_uri(), storage_name, db_type=stype, sql_echo=False)

        # Clean and re-init the database
        storage._clear_db(storage_name)
    else:
        raise KeyError("Storage type {} not understood".format(stype))

    yield storage

    if stype == "sqlalchemy":
        # todo: drop db
        # storage._clear_db(storage_name)
        pass
    else:
        raise KeyError("Storage type {} not understood".format(stype))


@pytest.fixture(scope="module", params=["sqlalchemy"])
def socket_fixture(request):

    yield from build_socket_fixture(request.param)


@pytest.fixture(scope="module")
def sqlalchemy_socket_fixture(request, postgres_server):

    yield from build_socket_fixture("sqlalchemy", postgres_server)


def live_fractal_or_skip():
    """
    Ensure Fractal live connection can be made
    First looks for a local staging server, then tries QCArchive.
    """
    try:
        return FractalClient("localhost:7777", verify=False)
    except (requests.exceptions.ConnectionError, ConnectionRefusedError):
        print("Failed to connect to localhost, trying MolSSI QCArchive.")
        try:
            requests.get("https://api.qcarchive.molssi.org:443", json={}, timeout=5)
            return FractalClient()
        except (requests.exceptions.ConnectionError, ConnectionRefusedError):
            return pytest.skip("Could not make a connection to central Fractal server")


def df_compare(df1, df2, sort=False):
    """ checks equality even when columns contain numpy arrays, which .equals and == struggle with """
    if sort:
        if isinstance(df1, pd.DataFrame):
            df1 = df1.reindex(sorted(df1.columns), axis=1)
        elif isinstance(df1, pd.Series):
            df1 = df1.sort_index()
        if isinstance(df2, pd.DataFrame):
            df2 = df2.reindex(sorted(df2.columns), axis=1)
        elif isinstance(df2, pd.Series):
            df2 = df2.sort_index()

    def element_equal(e1, e2):
        if isinstance(e1, np.ndarray):
            if not np.array_equal(e1, e2):
                return False
        elif isinstance(e1, Molecule):
            if not e1.get_hash() == e2.get_hash():
                return False
        # Because nan != nan
        elif isinstance(e1, float) and np.isnan(e1):
            if not np.isnan(e2):
                return False
        else:
            if not e1 == e2:
                return False
        return True

    if isinstance(df1, pd.Series):
        if not isinstance(df2, pd.Series):
            return False
        if len(df1) != len(df2):
            return False
        for i in range(len(df1)):
            if not element_equal(df1[i], df2[i]):
                return False
        return True

    for column in df1.columns:
        if column.startswith("_"):
            df1.drop(column, axis=1, inplace=True)
    for column in df2.columns:
        if column.startswith("_"):
            df2.drop(column, axis=1, inplace=True)
    if not all(df1.columns == df2.columns):
        return False
    if not all(df1.index.values == df2.index.values):
        return False
    for i in range(df1.shape[0]):
        for j in range(df1.shape[1]):
            if not element_equal(df1.iloc[i, j], df2.iloc[i, j]):
                return False

    return True



from typing import Any, Dict, Optional

from .gridoptimization import GridOptimizationRecord
from .records import OptimizationRecord, ResultRecord
from .torsiondrive import TorsionDriveRecord


def build_procedure(
    data: Dict[str, Any], procedure: Optional[str] = None, client: Optional["FractalClient"] = None
) -> "BaseRecord":
    """
    Constructs a Service ORM from incoming JSON data.

    Parameters
    ----------
    data : Dict[str, Any]
        A JSON representation of the procedure.
    procedure : Optional[str], optional
        The name of the procedure. If blank the procedure name is pulled from the `data["procedure"]` field.
    client : Optional['FractalClient'], optional
        A FractalClient connected to a server.

    Returns
    -------
    ret : BaseRecord
        Returns an interface object of the appropriate procedure.

    Examples
    --------

    # A partial example of torsiondrive metadata
    >>> data = {
        "procedure": "torsiondrive",
        "initial_molecule": "5b7f1fd57b87872d2c5d0a6c",
        "state": "RUNNING",
        "id": "5b7f1fd57b87872d2c5d0a6d",
        ....
    }

    >>> build_orm(data)
    TorsionDriveRecord(id='5b7f1fd57b87872d2c5d0a6c', state='RUNNING', molecule_id='5b7f1fd57b87872d2c5d0a6c', molecule_name='HOOH')
    """

    if ("procedure" not in data) and (procedure is None):
        raise KeyError("There is not a procedure tag and procedure is none. Unable to determine procedure type")

    # import json
    # print(json.dumps(data, indent=2))
    if data["procedure"].lower() == "single":
        return ResultRecord(**data, client=client)
    elif data["procedure"].lower() == "torsiondrive":
        return TorsionDriveRecord(**data, client=client)
    elif data["procedure"].lower() == "gridoptimization":
        return GridOptimizationRecord(**data, client=client)
    elif data["procedure"].lower() == "optimization":
        return OptimizationRecord(**data, client=client)
    else:
        raise KeyError("Service names {} not recognized.".format(data["procedure"]))



"""
Tests for QCFractals CLI
"""
import os
import tempfile
import time
from typing import Any, Dict

import pytest
import yaml

import qcfractal
from qcfractal import testing
from qcfractal.cli.cli_utils import read_config_file

_options = {"coverage": True, "dump_stdout": True}
_pwd = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture(scope="module")
def qcfractal_base_init():

    storage = qcfractal.TemporaryPostgres()
    tmpdir = tempfile.TemporaryDirectory()

    args = [
        "qcfractal-server",
        "init",
        "--base-folder",
        str(tmpdir.name),
        "--db-own=False",
        "--clear-database",
        f"--db-port={storage.config.database.port}",
    ]
    assert testing.run_process(args, **_options)

    yield f"--base-folder={tmpdir.name}"


@pytest.mark.slow
def test_cli_server_boot(qcfractal_base_init):
    port = "--port=" + str(testing.find_open_port())
    args = ["qcfractal-server", "start", qcfractal_base_init, port]
    assert testing.run_process(args, interupt_after=10, **_options)


@pytest.mark.slow
def test_cli_upgrade(qcfractal_base_init):
    args = ["qcfractal-server", "upgrade", qcfractal_base_init]
    assert testing.run_process(args, interupt_after=10, **_options)


@pytest.mark.slow
def test_cli_user_add(qcfractal_base_init):
    args = ["qcfractal-server", "user", qcfractal_base_init, "add", "test_user_add_1", "--permissions", "admin"]
    assert testing.run_process(args, **_options)

    args = ["qcfractal-server", "user", qcfractal_base_init, "add", "test_user_add_1", "--permissions", "admin"]
    assert testing.run_process(args, **_options) is False

    args = [
        "qcfractal-server",
        "user",
        qcfractal_base_init,
        "add",
        "test_user_add_2",
        "--password",
        "foo",
        "--permissions",
        "admin",
    ]
    assert testing.run_process(args, **_options)

    args = ["qcfractal-server", "user", qcfractal_base_init, "add", "test_user_add_3"]
    assert testing.run_process(args, **_options) is False


@pytest.mark.slow
def test_cli_user_show(qcfractal_base_init):
    args = ["qcfractal-server", "user", qcfractal_base_init, "add", "test_user_show", "--permissions", "admin"]
    assert testing.run_process(args, **_options)

    args = ["qcfractal-server", "user", qcfractal_base_init, "info", "test_user_show"]
    assert testing.run_process(args, **_options)

    args = ["qcfractal-server", "user", qcfractal_base_init, "info", "badname_1234"]
    assert testing.run_process(args, **_options) is False


@pytest.mark.slow
def test_cli_user_modify(qcfractal_base_init):
    args = ["qcfractal-server", "user", qcfractal_base_init, "add", "test_user_modify", "--permissions", "read"]
    assert testing.run_process(args, **_options)

    args = [
        "qcfractal-server",
        "user",
        qcfractal_base_init,
        "modify",
        "test_user_modify",
        "--permissions",
        "read",
        "write",
        "--reset-password",
    ]
    assert testing.run_process(args, **_options)

    args = ["qcfractal-server", "user", qcfractal_base_init, "modify", "test_user_modify", "--password", "foopass"]
    assert testing.run_process(args, **_options)

    args = ["qcfractal-server", "user", qcfractal_base_init, "modify", "test_user_modify", "--permissions", "read"]
    assert testing.run_process(args, **_options)

    args = ["qcfractal-server", "user", qcfractal_base_init, "modify", "badname_1234"]
    assert testing.run_process(args, **_options) is False


@pytest.mark.slow
def test_cli_user_remove(qcfractal_base_init):
    args = ["qcfractal-server", "user", qcfractal_base_init, "add", "test_user_remove", "--permissions", "admin"]
    assert testing.run_process(args, **_options)

    args = ["qcfractal-server", "user", qcfractal_base_init, "remove", "test_user_remove"]
    assert testing.run_process(args, **_options)

    args = ["qcfractal-server", "user", qcfractal_base_init, "remove", "badname_1234"]
    assert testing.run_process(args, **_options) is False


@pytest.mark.xfail(reason="Failing on Travis for unknown reasons.")
@pytest.mark.slow
def test_cli_server_local_boot(qcfractal_base_init):
    port = "--port=" + str(testing.find_open_port())
    args = ["qcfractal-server", "start", "--local-manager=1", port, qcfractal_base_init]
    assert testing.run_process(args, interupt_after=10, **_options)


@pytest.fixture(scope="module")
def active_server(request, qcfractal_base_init):
    port = str(testing.find_open_port())
    args = ["qcfractal-server", "start", qcfractal_base_init, f"--port={port}"]
    assert testing.run_process(args, interupt_after=10, **_options)
    with testing.popen(args, **_options) as server:
        time.sleep(2)

        server.test_uri_cli = "--fractal-uri=localhost:" + port
        yield server


@pytest.mark.slow
@pytest.mark.parametrize("log_apis", [0, 1])
def test_with_api_logging(postgres_server, log_apis):

    tmpdir = tempfile.TemporaryDirectory()

    args = [
        "qcfractal-server",
        "init",
        "--base-folder",
        str(tmpdir.name),
        "--db-own=False",
        "--clear-database",
        f"--db-port={postgres_server.config.database.port}",
        f"--log-apis={log_apis}",
    ]
    assert testing.run_process(args, **_options)

    port = "--port=" + str(testing.find_open_port())
    args = ["qcfractal-server", "start", f"--base-folder={tmpdir.name}", port]
    assert testing.run_process(args, interupt_after=10, **_options)


@pytest.mark.slow
def test_manager_local_testing_process():
    assert testing.run_process(["qcfractal-manager", "--adapter=pool", "--test", "--tasks-per-worker=2"], **_options)


@pytest.mark.slow
def test_manager_executor_manager_boot(active_server):
    args = ["qcfractal-manager", active_server.test_uri_cli, "--adapter=pool", "--tasks-per-worker=2", "--verify=False"]
    assert testing.run_process(args, interupt_after=7, **_options)


@pytest.mark.slow
def test_manager_executor_manager_boot_from_file(active_server, tmp_path):

    yaml_file = """
    common:
        adapter: pool
        tasks_per_worker: 4
        cores_per_worker: 4
    server:
        fractal_uri: {}
        verify: False
    """.format(
        active_server.test_uri_cli.split("=")[1]
    )

    p = tmp_path / "config.yaml"
    p.write_text(yaml_file)

    args = ["qcfractal-manager", "--config-file={}".format(p)]
    assert testing.run_process(args, interupt_after=7, **_options)


@pytest.mark.slow
def cli_manager_runs(config_data, tmp_path):
    temp_config = tmp_path / "temp_config.yaml"
    temp_config.write_text(yaml.dump(config_data))
    args = ["qcfractal-manager", f"--config-file={temp_config}", "--test"]
    assert testing.run_process(args, **_options)


@pytest.mark.slow
def load_manager_config(adapter, scheduler) -> Dict[str, Any]:
    config = read_config_file(os.path.join(_pwd, "manager_boot_template.yaml"))
    config["common"]["adapter"] = adapter
    config["cluster"]["scheduler"] = scheduler
    return config


@pytest.mark.slow
@pytest.mark.parametrize(
    "adapter,scheduler",
    [
        ("pool", "slurm"),
        pytest.param("dask", "slurm", marks=testing.using_dask_jobqueue),
        pytest.param("dask", "PBS", marks=testing.using_dask_jobqueue),
        pytest.param("dask", "MoAb", marks=testing.using_dask_jobqueue),
        pytest.param("dask", "SGE", marks=testing.using_dask_jobqueue),
        pytest.param("dask", "lSf", marks=testing.using_dask_jobqueue),
        pytest.param("parsl", "slurm", marks=testing.using_parsl),
        pytest.param("parsl", "PBS", marks=testing.using_parsl),
        pytest.param("parsl", "MoAb", marks=testing.using_parsl),
        pytest.param("parsl", "SGE", marks=testing.using_parsl),
        pytest.param("parsl", "lSf", marks=[testing.using_parsl, pytest.mark.xfail]),  # Invalid combination
        pytest.param("NotAParser", "slurm", marks=pytest.mark.xfail),  # Invalid Parser
        pytest.param("pool", "NotAScheduler", marks=pytest.mark.xfail),  # Invalid Scheduler
    ],
)
def test_cli_managers(adapter, scheduler, tmp_path):
    """Test that multiple adapter/scheduler combinations at least can boot up in Managers"""
    config = load_manager_config(adapter, scheduler)
    cli_manager_runs(config, tmp_path)


@pytest.mark.slow
@testing.using_parsl
def test_cli_manager_parsl_launchers(tmp_path):
    config = load_manager_config("parsl", "slurm")
    config["parsl"]["provider"].update({"launcher": {"launcher_class": "singleNODELauncher"}})
    cli_manager_runs(config, tmp_path)


@pytest.mark.slow
@pytest.mark.parametrize(
    "adapter",
    [pytest.param("dask", marks=testing.using_dask_jobqueue), pytest.param("parsl", marks=testing.using_parsl)],
)
def test_cli_managers_missing(adapter, tmp_path):
    """Test that the manager block missing correctly sets defaults"""
    config = load_manager_config(adapter, "slurm")
    config.pop(adapter, None)
    cli_manager_runs(config, tmp_path)


@pytest.mark.slow
@pytest.mark.parametrize(
    "adapter",
    [pytest.param("dask", marks=testing.using_dask_jobqueue), pytest.param("parsl", marks=testing.using_parsl)],
)
def test_cli_managers_none(adapter, tmp_path):
    """Test that manager block set to None correctly assigns the defaults"""
    config = load_manager_config(adapter, "slurm")
    config[adapter] = None
    cli_manager_runs(config, tmp_path)


@pytest.mark.slow
def test_cli_managers_help():
    """Test that qcfractal_manager --help works"""
    args = ["qcfractal-manager", "--help"]
    testing.run_process(args, **_options)


@pytest.mark.slow
def test_cli_managers_schema():
    """Test that qcfractal_manager --schema works"""
    args = ["qcfractal-manager", "--schema"]
    testing.run_process(args, **_options)


@pytest.mark.slow
def test_cli_managers_skel(tmp_path):
    """Test that qcfractal_manager --skeleton works"""
    config = tmp_path / "config.yaml"
    args = ["qcfractal-manager", "--skel", config.as_posix()]
    testing.run_process(args, **_options)


@testing.using_parsl
def test_nodeparallel_tasks(tmp_path):
    """Make sure that it boots up properly"""
    config = load_manager_config("parsl", "cobalt")
    config["common"]["nodes_per_task"] = 2
    config["common"]["nodes_per_job"] = 2
    config["common"]["cores_per_rank"] = 2
    cli_manager_runs(config, tmp_path)



"""
The global qcfractal config file specification.
"""

import argparse
import os
from pathlib import Path
from typing import Optional

import yaml
from pydantic import Field, validator

from .interface.models import AutodocBaseSettings


def _str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


class SettingsCommonConfig:
    env_prefix = "QCF_"
    case_insensitive = True
    extra = "forbid"


class ConfigSettings(AutodocBaseSettings):

    _type_map = {"string": str, "integer": int, "float": float, "boolean": _str2bool}

    @classmethod
    def field_names(cls):
        return list(cls.schema()["properties"].keys())

    @classmethod
    def help_info(cls, field):
        info = cls.schema()["properties"][field]

        ret = {"type": cls._type_map[info["type"]]}
        # if "default" in info:
        #     ret["default"] = info["default"]
        if "description" in info:
            ret["help"] = info["description"]
        return ret


class DatabaseSettings(ConfigSettings):
    """
    Postgres Database settings
    """

    port: int = Field(5432, description="The postgresql default port")
    host: str = Field(
        "localhost",
        description="Default location for the postgres server. If not localhost, qcfractal command lines cannot manage "
        "the instance.",
    )
    username: str = Field(None, description="The postgres username to default to.")
    password: str = Field(None, description="The postgres password for the give user.")
    directory: str = Field(
        None, description="The physical location of the QCFractal instance data, defaults to the root folder."
    )
    database_name: str = Field("qcfractal_default", description="The database name to connect to.")
    logfile: str = Field("qcfractal_postgres.log", description="The logfile to write postgres logs.")
    own: bool = Field(
        True,
        description="If own is True, QCFractal will control the database instance. If False "
        "Postgres will expect a booted server at the database specification.",
    )

    class Config(SettingsCommonConfig):
        pass


class ViewSettings(ConfigSettings):
    """
    HDF5 view settings
    """

    enable: bool = Field(True, description="Enable frozen-views.")
    directory: str = Field(None, description="Location of frozen-view data. If None, defaults to base_folder/views.")


class FractalServerSettings(ConfigSettings):
    """
    Fractal Server settings
    """

    name: str = Field("QCFractal Server", description="The QCFractal server default name.")
    port: int = Field(7777, description="The QCFractal default port.")

    compress_response: bool = Field(
        True, description="Compress REST responses or not, should be True unless behind a proxy."
    )
    allow_read: bool = Field(True, description="Always allows read access to record tables.")
    security: str = Field(
        None,
        description="Optional user authentication. Specify 'local' to enable "
        "authentication through locally stored usernames. "
        "User permissions may be manipulated through the ``qcfractal-server "
        "user`` CLI.",
    )

    query_limit: int = Field(1000, description="The maximum number of records to return per query.")
    logfile: Optional[str] = Field("qcfractal_server.log", description="The logfile to write server logs.")
    loglevel: str = Field("info", description="Level of logging to enable (debug, info, warning, error, critical)")
    cprofile: Optional[str] = Field(
        None, description="Enable profiling via cProfile, and output cprofile data to this path"
    )
    service_frequency: int = Field(60, description="The frequency to update the QCFractal services.")
    max_active_services: int = Field(20, description="The maximum number of concurrent active services.")
    heartbeat_frequency: int = Field(1800, description="The frequency (in seconds) to check the heartbeat of workers.")
    log_apis: bool = Field(
        False,
        description="True or False. Store API access in the Database. This is an advanced "
        "option for servers accessed by external users through QCPortal.",
    )
    geo_file_path: Optional[str] = Field(
        None,
        description="Geoip2 cites file path (.mmdb) for resolving IP addresses. Defaults to [base_folder]/GeoLite2-City.mmdb",
    )

    _default_geo_filename: str = "GeoLite2-City.mmdb"

    @validator("logfile")
    def check_basis(cls, v):
        if v == "None":
            v = None
        return v

    class Config(SettingsCommonConfig):
        pass


class FractalConfig(ConfigSettings):
    """
    Top level configuration headers and options for a QCFractal Configuration File
    """

    # class variable, not in the pydantic model
    _defaults_file_path: str = os.path.expanduser("~/.qca/qcfractal_defaults.yaml")

    base_folder: str = Field(
        os.path.expanduser("~/.qca/qcfractal"),
        description="The QCFractal base instance to attach to. " "Default will be your home directory",
    )
    database: DatabaseSettings = DatabaseSettings()
    view: ViewSettings = ViewSettings()
    fractal: FractalServerSettings = FractalServerSettings()

    class Config(SettingsCommonConfig):
        pass

    def __init__(self, **kwargs):

        # If no base_folder provided, read it from ~/.qca/qcfractal_defaults.yaml (if it exists)
        # else, use the default base_folder
        if "base_folder" in kwargs:
            kwargs["base_folder"] = os.path.expanduser(kwargs["base_folder"])
        else:
            if Path(FractalConfig._defaults_file_path).exists():
                with open(FractalConfig._defaults_file_path, "r") as handle:
                    kwargs["base_folder"] = yaml.load(handle.read(), Loader=yaml.FullLoader)["default_base_folder"]

        super().__init__(**kwargs)

    @classmethod
    def from_base_folder(cls, base_folder):
        path = Path(base_folder).absolute() / "qcfractal_config.yaml"
        with open(str(path), "r") as handle:
            return cls(**yaml.load(handle.read(), Loader=yaml.FullLoader))

    @property
    def base_path(self):
        return Path(self.base_folder)

    @property
    def config_file_path(self):
        return self.base_path / "qcfractal_config.yaml"

    @property
    def database_path(self):
        if self.database.directory is None:
            return self.base_path / "postgres"
        else:
            return Path(os.path.expanduser(self.database.directory))

    def database_uri(self, safe: bool = True, database: str = None) -> str:

        uri = "postgresql://"
        if self.database.username is not None:
            uri += f"{self.database.username}:"

            if self.database.password is not None:
                if safe:
                    pw = "*******"
                else:
                    pw = self.database.password
                uri += pw

            uri += "@"

        uri += f"{self.database.host}:{self.database.port}/"

        if database is None:
            uri += self.database.database_name
        else:
            uri += database

        return uri

    @property
    def view_path(self):
        if self.view.directory is None:
            default_view_path = self.base_path / "views"
            default_view_path.mkdir(parents=False, exist_ok=True)
            return default_view_path
        else:
            return Path(os.path.expanduser(self.view.directory))

    def geo_file_path(self):

        if self.fractal.geo_file_path:
            return self.fractal.geo_file_path
        else:
            return os.path.join(self.base_folder, self.fractal._default_geo_filename)



"""
Either pull in QCEl models or local models
"""

from . import rest_models
from .common_models import (
    AutodocBaseSettings,
    Citation,
    KeywordSet,
    CompressionEnum,
    KVStore,
    Molecule,
    ObjectId,
    OptimizationProtocols,
    OptimizationSpecification,
    ProtoModel,
    QCSpecification,
    ResultProtocols,
)
from .gridoptimization import GridOptimizationInput, GridOptimizationRecord
from .model_builder import build_procedure
from .model_utils import hash_dictionary, json_encoders, prepare_basis
from .records import OptimizationRecord, ResultRecord
from .rest_models import ComputeResponse, rest_model
from .task_models import ManagerStatusEnum, PythonComputeSpec, TaskRecord, TaskStatusEnum
from .torsiondrive import TorsionDriveInput, TorsionDriveRecord



"""Compression of KVStore

Revision ID: 30fd8253d87f
Revises: 1604623c481a
Create Date: 2020-08-30 10:11:57.574292

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "30fd8253d87f"
down_revision = "1604623c481a"
branch_labels = None
depends_on = None


def upgrade():
    # Appears there is an alembic issue with autogenerating enums:
    # https://github.com/sqlalchemy/alembic/issues/278

    compression_enum = postgresql.ENUM("none", "gzip", "bzip2", "lzma", name="compressionenum")
    compression_enum.create(op.get_bind())

    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("kv_store", sa.Column("compression", compression_enum, nullable=True))
    op.add_column("kv_store", sa.Column("compression_level", sa.Integer(), nullable=True))
    op.add_column("kv_store", sa.Column("data", sa.LargeBinary(), nullable=True))
    op.alter_column("kv_store", "value", existing_type=postgresql.JSON(astext_type=sa.Text()), nullable=True)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("kv_store", "value", existing_type=postgresql.JSON(astext_type=sa.Text()), nullable=False)
    op.drop_column("kv_store", "data")
    op.drop_column("kv_store", "compression_level")
    op.drop_column("kv_store", "compression")
    op.execute("DROP TYPE compressionenum")
    # ### end Alembic commands ###



import datetime
from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import Field, validator
from qcelemental.models import ComputeError

from .common_models import ObjectId, ProtoModel


class TaskStatusEnum(str, Enum):
    """
    The state of a Task object. The states which are available are a finite set.
    """

    running = "RUNNING"
    waiting = "WAITING"
    error = "ERROR"
    complete = "COMPLETE"


class ManagerStatusEnum(str, Enum):
    """
    The state of a Queue Manager. The states which are available are a finite set.
    """

    active = "ACTIVE"
    inactive = "INACTIVE"


class PriorityEnum(int, Enum):
    """
    The priority of a Task. Higher priority will be pulled first. The priorities which are available are a finite set.
    """

    HIGH = 2
    NORMAL = 1
    LOW = 0


class BaseResultEnum(str, Enum):
    result = "result"
    procedure = "procedure"


class PythonComputeSpec(ProtoModel):
    function: str = Field(
        ..., description="The module and function name of a Python-callable to call. Of the form 'module.function'."
    )
    args: List[Any] = Field(
        ..., description="A List of positional arguments to pass into ``function`` in order they appear."
    )
    kwargs: Dict[str, Any] = Field(..., description="Dictionary of keyword arguments to pass into ``function``.")


class TaskRecord(ProtoModel):

    id: ObjectId = Field(None, description="The Database assigned Id of the Task, if it has been assigned yet.")

    spec: PythonComputeSpec = Field(..., description="The Python function specification for this Task.")
    parser: str = Field(..., description="The type of operation this is Task is. Can be 'single' or 'optimization'.")
    status: TaskStatusEnum = Field(TaskStatusEnum.waiting, description="What stage of processing this task is at.")

    # Compute blockers and prevention
    program: str = Field(
        ..., description="Name of the quantum chemistry program which must be present to execute this task."
    )
    procedure: Optional[str] = Field(
        None, description="Name of the procedure the compute platform must be able to perform to execute this task."
    )
    manager: Optional[str] = Field(None, description="The Queue Manager that evaluated this task.")

    # Sortables
    priority: PriorityEnum = Field(PriorityEnum.NORMAL, description=str(PriorityEnum.__doc__))
    tag: Optional[str] = Field(
        None,
        description="The optional tag assigned to this Task. Tagged tasks can only be pulled by Queue Managers which "
        "explicitly reference this tag. If no Tag is specified, any Queue Manager can pull this Task.",
    )
    # Link back to the base Result
    base_result: ObjectId = Field(
        ..., description="Reference to the output Result from this Task as it exists within the database."
    )

    # Modified data
    modified_on: datetime.datetime = Field(None, description="The last time this task was updated in the Database.")
    created_on: datetime.datetime = Field(None, description="The time when this task was created in the Database.")

    def __init__(self, **data):

        # Set datetime defaults if not present
        dt = datetime.datetime.utcnow()
        data.setdefault("modified_on", dt)
        data.setdefault("created_on", dt)

        super().__init__(**data)

    @validator("priority", pre=True)
    def munge_priority(cls, v):
        if isinstance(v, str):
            v = PriorityEnum[v.upper()]
        elif v is None:
            v = PriorityEnum.NORMAL
        return v

    @validator("program")
    def check_program(cls, v):
        return v.lower()

    @validator("procedure")
    def check_procedure(cls, v):
        if v:
            v = v.lower()
        return v



from typing import List, Optional, Set, Union

from sqlalchemy import Integer, inspect
from sqlalchemy.sql import bindparam, text

from qcfractal.interface.models import Molecule, ResultRecord
from qcfractal.storage_sockets.models import MoleculeORM, ResultORM

QUERY_CLASSES = set()


class QueryBase:

    # The name/alias used by the REST APIs to access this class
    _class_name = None
    _available_groupby = set()

    # Mapping of the requested feature and the internal query method
    _query_method_map = {}

    def __init__(self, database_name, max_limit=1000):
        self.database_name = database_name
        self.max_limit = max_limit

    def __init_subclass__(cls, **kwargs):
        if cls not in QUERY_CLASSES:
            QUERY_CLASSES.add(cls)
        super().__init_subclass__(**kwargs)

    def query(self, session, query_key, limit=0, skip=0, include=None, exclude=None, **kwargs):

        if query_key not in self._query_method_map:
            raise TypeError(f"Query type {query_key} is unimplemented for class {self._class_name}")

        self.session = session

        return getattr(self, self._query_method_map[query_key])(**kwargs)

    def execute_query(self, sql_statement, with_keys=True, **kwargs):
        """Execute sql statemet, apply limit, and return results as dict if needed"""

        # TODO: check count first, way to iterate

        # sql_statement += f' LIMIT {self.max_limit}'
        result = self.session.execute(sql_statement, kwargs)
        keys = result.keys()  # get keys before fetching
        result = result.fetchall()
        self.session.commit()

        # create a list of dict with the keys and values of the results (instead of tuples)
        if with_keys:
            result = [dict(zip(keys, res)) for res in result]

        return result

    def _base_count(self, table_name: str, available_groupbys: Set[str], groupby: Optional[List[str]] = None):
        if groupby:
            bad_groups = set(groupby) - available_groupbys
            if bad_groups:
                raise AttributeError(f"The following groups are not permissible: {missing}")

            global_str = ", ".join(groupby)
            select_str = global_str + ", "
            extra_str = f"""GROUP BY {global_str}\nORDER BY {global_str}"""

        else:
            select_str = ""
            extra_str = ""

        sql_statement = f"""
select {select_str}count(*) from {table_name}
{extra_str}
"""

        ret = self.execute_query(sql_statement, with_keys=True)

        if groupby:
            return ret
        else:
            return ret[0]["count"]

    @staticmethod
    def _raise_missing_attribute(cls, query_key, missing_attribute, amend_msg=""):
        """Raises error for missing attribute in a message suitable for the REST user"""

        raise AttributeError(f"To query {cls._class_name} for {query_key} " f"you must provide {missing_attribute}.")


# ----------------------------------------------------------------------------


class TaskQueries(QueryBase):

    _class_name = "task"
    _query_method_map = {"counts": "_task_counts"}

    def _task_counts(self):

        sql_statement = f"""
            SELECT tag, priority, status, count(*)
            FROM task_queue
            WHERE True
            group by tag, priority, status
            order by tag, priority, status
        """

        return self.execute_query(sql_statement, with_keys=True)


# ----------------------------------------------------------------------------


class DatabaseStatQueries(QueryBase):

    _class_name = "database_stats"

    _query_method_map = {
        "table_count": "_table_count",
        "database_size": "_database_size",
        "table_information": "_table_information",
    }

    def _table_count(self, table_name=None):

        if table_name is None:
            self._raise_missing_attribute("table_name", "table name")

        sql_statement = f"SELECT count(*) from {table_name}"
        return self.execute_query(sql_statement, with_keys=False)[0]

    def _database_size(self):

        sql_statement = f"SELECT pg_database_size('{self.database_name}')"
        return self.execute_query(sql_statement, with_keys=True)[0]["pg_database_size"]

    def _table_information(self):

        sql_statement = f"""
SELECT relname                                AS table_name
     , c.reltuples::BIGINT                    AS row_estimate
     , pg_total_relation_size(c.oid)          AS total_bytes
     , pg_indexes_size(c.oid)                 AS index_bytes
     , pg_total_relation_size(reltoastrelid)  AS toast_bytes
FROM pg_class c
         LEFT JOIN pg_namespace n ON n.oid = c.relnamespace
WHERE relkind = 'r';
 """

        result = self.execute_query(sql_statement, with_keys=False)

        ret = []
        for row in result:
            if ("pg_" in row[0]) or ("sql_" in row[0]):
                continue
            ret.append(list(row))

        ret = {"columns": ["table_name", "row_estimate", "total_bytes", "index_bytes", "toast_bytes"], "rows": ret}

        return ret


class ResultQueries(QueryBase):

    _class_name = "result"

    _query_method_map = {"count": "_count"}

    def _count(self, groupby: Optional[List[str]] = None):

        available_groupbys = {"result_type", "status"}

        return self._base_count("base_result", available_groupbys, groupby=groupby)


class MoleculeQueries(QueryBase):

    _class_name = "molecule"

    _query_method_map = {"count": "_count"}

    def _count(self, groupby: Optional[List[str]] = None):

        available_groupbys = set()

        return self._base_count("molecule", available_groupbys, groupby=groupby)


# ----------------------------------------------------------------------------


class TorsionDriveQueries(QueryBase):

    _class_name = "torsiondrive"

    _query_method_map = {
        "initial_molecules": "_get_initial_molecules",
        "initial_molecules_ids": "_get_initial_molecules_ids",
        "final_molecules": "_get_final_molecules",
        "final_molecules_ids": "_get_final_molecules_ids",
        "return_results": "_get_return_results",
    }

    def _get_initial_molecules_ids(self, torsion_id=None):

        if torsion_id is None:
            self._raise_missing_attribute("initial_molecules_ids", "torsion drive id")

        sql_statement = f"""
                select initial_molecule from optimization_procedure as opt where opt.id in
                (
                    select opt_id from optimization_history where torsion_id = {torsion_id}
                )
                order by opt.id
        """

        return self.execute_query(sql_statement, with_keys=False)

    def _get_initial_molecules(self, torsion_id=None):

        if torsion_id is None:
            self._raise_missing_attribute("initial_molecules", "torsion drive id")

        sql_statement = f"""
                select molecule.* from molecule
                join optimization_procedure as opt
                on molecule.id = opt.initial_molecule
                where opt.id in
                    (select opt_id from optimization_history where torsion_id = {torsion_id})
        """

        return self.execute_query(sql_statement, with_keys=True)

    def _get_final_molecules_ids(self, torsion_id=None):

        if torsion_id is None:
            self._raise_missing_attribute("final_molecules_ids", "torsion drive id")

        sql_statement = f"""
                select final_molecule from optimization_procedure as opt where opt.id in
                (
                    select opt_id from optimization_history where torsion_id = {torsion_id}
                )
                order by opt.id
        """

        return self.execute_query(sql_statement, with_keys=False)

    def _get_final_molecules(self, torsion_id=None):

        if torsion_id is None:
            self._raise_missing_attribute("final_molecules", "torsion drive id")

        sql_statement = f"""
                select molecule.* from molecule
                join optimization_procedure as opt
                on molecule.id = opt.final_molecule
                where opt.id in
                    (select opt_id from optimization_history where torsion_id = {torsion_id})
        """

        return self.execute_query(sql_statement, with_keys=True)

    def _get_return_results(self, torsion_id=None):
        """All return results ids of a torsion drive"""

        if torsion_id is None:
            self._raise_missing_attribute("return_results", "torsion drive id")

        sql_statement = f"""
                select opt_res.opt_id, result.id as result_id, result.return_result from result
                join opt_result_association as opt_res
                on result.id = opt_res.result_id
                where opt_res.opt_id in
                (
                    select opt_id from optimization_history where torsion_id = {torsion_id}
                )
        """

        return self.execute_query(sql_statement, with_keys=False)


class OptimizationQueries(QueryBase):

    _class_name = "optimization"
    _exclude = ["molecule_hash", "molecular_formula", "result_type"]
    _query_method_map = {
        "all_results": "_get_all_results",
        "final_result": "_get_final_results",
        "initial_molecule": "_get_initial_molecules",
        "final_molecule": "_get_final_molecules",
    }

    def _remove_excluded_keys(self, data):
        for key in self._exclude:
            data.pop(key, None)

    def _get_all_results(self, optimization_ids: List[Union[int, str]] = None):
        """Returns all the results objects (trajectory) of each optmization
        Returns list(list)"""

        if optimization_ids is None:
            self._raise_missing_attribute("all_results", "List of optimizations ids")

        # row_to_json(result.*)
        sql_statement = text(
            """
            select * from base_result
            join (
                select opt_id, result.* from result
                join opt_result_association as traj
                on result.id = traj.result_id
                where traj.opt_id in :optimization_ids
            ) result
            on base_result.id = result.id
        """
        )

        # bind and expand ids list
        sql_statement = sql_statement.bindparams(bindparam("optimization_ids", expanding=True))

        # column types:
        columns = inspect(ResultORM).columns
        sql_statement = sql_statement.columns(opt_id=Integer, *columns)
        query_result = self.execute_query(sql_statement, optimization_ids=list(optimization_ids))

        ret = {}
        for rec in query_result:
            self._remove_excluded_keys(rec)
            key = rec.pop("opt_id")
            if key not in ret:
                ret[key] = []

            ret[key].append(ResultRecord(**rec))

        return ret

    def _get_final_results(self, optimization_ids: List[Union[int, str]] = None):
        """Return the actual results objects of the best result in each optimization"""

        if optimization_ids is None:
            self._raise_missing_attribute("final_result", "List of optimizations ids")

        sql_statement = text(
            """
            select * from base_result
            join (
                select opt_id, result.* from result
                join (
                    select opt.opt_id, opt.result_id, max_pos from opt_result_association as opt
                    inner join (
                            select opt_id, max(position) as max_pos from opt_result_association
                            where opt_id in :optimization_ids
                            group by opt_id
                        ) opt2
                    on opt.opt_id = opt2.opt_id and opt.position = opt2.max_pos
                ) traj
                on result.id = traj.result_id
            ) result
            on base_result.id = result.id
        """
        )

        # bind and expand ids list
        sql_statement = sql_statement.bindparams(bindparam("optimization_ids", expanding=True))

        # column types:
        columns = inspect(ResultORM).columns
        sql_statement = sql_statement.columns(opt_id=Integer, *columns)
        query_result = self.execute_query(sql_statement, optimization_ids=list(optimization_ids))

        ret = {}
        for rec in query_result:
            self._remove_excluded_keys(rec)
            key = rec.pop("opt_id")
            ret[key] = ResultRecord(**rec)

        return ret

    def _get_initial_molecules(self, optimization_ids=None):

        if optimization_ids is None:
            self._raise_missing_attribute("initial_molecule", "List of optimizations ids")

        sql_statement = text(
            """
                select opt.id as opt_id, molecule.* from molecule
                join optimization_procedure as opt
                on molecule.id = opt.initial_molecule
                where opt.id in :optimization_ids
        """
        )

        # bind and expand ids list
        sql_statement = sql_statement.bindparams(bindparam("optimization_ids", expanding=True))

        # column types:
        columns = inspect(MoleculeORM).columns
        sql_statement = sql_statement.columns(opt_id=Integer, *columns)
        query_result = self.execute_query(sql_statement, optimization_ids=list(optimization_ids))

        ret = {}
        for rec in query_result:
            self._remove_excluded_keys(rec)
            key = rec.pop("opt_id")
            rec = {k: v for k, v in rec.items() if v is not None}
            ret[key] = Molecule(**rec)

        return ret

    def _get_final_molecules(self, optimization_ids=None):

        if optimization_ids is None:
            self._raise_missing_attribute("final_molecule", "List of optimizations ids")

        sql_statement = text(
            """
                select opt.id as opt_id, molecule.* from molecule
                join optimization_procedure as opt
                on molecule.id = opt.final_molecule
                where opt.id in :optimization_ids
        """
        )

        # bind and expand ids list
        sql_statement = sql_statement.bindparams(bindparam("optimization_ids", expanding=True))

        # column types:
        columns = inspect(MoleculeORM).columns
        sql_statement = sql_statement.columns(opt_id=Integer, *columns)
        query_result = self.execute_query(sql_statement, optimization_ids=list(optimization_ids))

        ret = {}
        for rec in query_result:
            self._remove_excluded_keys(rec)
            key = rec.pop("opt_id")
            rec = {k: v for k, v in rec.items() if v is not None}
            ret[key] = Molecule(**rec)

        return ret



"""Updates the Dataset primary keys.

Revision ID: cc54842bb6ba
Revises: 129ff3ce9247
Create Date: 2019-09-19 15:10:26.146940

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "cc54842bb6ba"
down_revision = "129ff3ce9247"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("ALTER TABLE dataset_entry DROP CONSTRAINT dataset_entry_pkey")
    op.create_primary_key(None, "dataset_entry", ["dataset_id", "name"])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("ALTER TABLE dataset_entry DROP CONSTRAINT dataset_entry_pkey")
    op.create_primary_key(None, "dataset_entry", ["dataset_id", "molecule_id"])
    # ### end Alembic commands ###



"""
A file containing utility functions for testing portal
"""


def compare_lists(bench, val):
    """
    Checks to see if two list like objects are the same.
    """

    if not len(bench) == len(val):
        return False
    if not sorted(bench) == sorted(val):
        return False

    return True



"""
Tests for the interface utility functions.
"""

from . import portal


def test_replace_dict_keys():

    ret = portal.util.replace_dict_keys({5: 5}, {5: 10})
    assert ret == {10: 5}

    ret = portal.util.replace_dict_keys({5: 5}, {10: 5})
    assert ret == {5: 5}

    ret = portal.util.replace_dict_keys([{5: 5}], {10: 5})
    assert ret == [{5: 5}]

    ret = portal.util.replace_dict_keys({5: {5: 10}}, {5: 10})
    assert ret == {10: {10: 10}}



"""
A model for TorsionDrive
"""

import copy
import json
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from pydantic import Field, constr, validator
from qcelemental import constants

from ..visualization import scatter_plot
from .common_models import Molecule, ObjectId, OptimizationSpecification, ProtoModel, QCSpecification
from .model_utils import recursive_normalizer
from .records import RecordBase

__all__ = ["TDKeywords", "TorsionDriveInput", "TorsionDriveRecord"]


class TDKeywords(ProtoModel):
    """
    TorsionDriveRecord options
    """

    dihedrals: List[Tuple[int, int, int, int]] = Field(
        ...,
        description="The list of dihedrals to select for the TorsionDrive operation. Each entry is a tuple of integers "
        "of for particle indices.",
    )
    grid_spacing: List[int] = Field(
        ...,
        description="List of grid spacing for dihedral scan in degrees. Multiple values will be mapped to each "
        "dihedral angle.",
    )
    dihedral_ranges: Optional[List[Tuple[int, int]]] = Field(
        None,
        description="A list of dihedral range limits as a pair (lower, upper). "
        "Each range corresponds to the dihedrals in input.",
    )
    energy_decrease_thresh: Optional[float] = Field(
        None,
        description="The threshold of the smallest energy decrease amount to trigger activating optimizations from "
        "grid point.",
    )
    energy_upper_limit: Optional[float] = Field(
        None,
        description="The threshold if the energy of a grid point that is higher than the current global minimum, to "
        "start new optimizations, in unit of a.u. I.e. if energy_upper_limit = 0.05, current global "
        "minimum energy is -9.9 , then a new task starting with energy -9.8 will be skipped.",
    )

    def __init__(self, **kwargs):
        super().__init__(**recursive_normalizer(kwargs))


_td_constr = constr(strip_whitespace=True, regex="torsiondrive")
_qcfractal_constr = constr(strip_whitespace=True, regex="qcfractal")


class TorsionDriveInput(ProtoModel):
    """
    A TorsionDriveRecord Input base class
    """

    program: _td_constr = Field(
        "torsiondrive",
        description="The name of the program. Fixed to 'torsiondrive' since this input model is only valid for it.",
    )
    procedure: _td_constr = Field(
        "torsiondrive",
        description="The name of the Procedure. Fixed to 'torsiondrive' since this input model is only valid for it.",
    )
    initial_molecule: List[Union[ObjectId, Molecule]] = Field(
        ...,
        description="The Molecule(s) to begin the TorsionDrive with. This can either be an existing Molecule in "
        "the database (through its :class:`ObjectId`) or a fully specified :class:`Molecule` model.",
    )
    keywords: TDKeywords = Field(
        ..., description="TorsionDrive-specific input arguments to pass into the TorsionDrive Procedure"
    )
    optimization_spec: OptimizationSpecification = Field(
        ...,
        description="The settings which describe how to conduct the energy optimizations at each step of the torsion "
        "scan.",
    )
    qc_spec: QCSpecification = Field(
        ...,
        description="The settings which describe the individual quantum chemistry calculations at each step of the "
        "optimization.",
    )

    @validator("initial_molecule", pre=True)
    def check_initial_molecules(cls, v):
        if isinstance(v, (str, dict, Molecule)):
            v = [v]
        return v


class TorsionDriveRecord(RecordBase):
    """
    A interface to the raw JSON data of a TorsionDriveRecord torsion scan run.
    """

    # Class data
    _hash_indices = {"initial_molecule", "keywords", "optimization_spec", "qc_spec"}

    # Version data
    version: int = Field(1, description="The version number of the Record.")
    procedure: _td_constr = Field(
        "torsiondrive",
        description="The name of the procedure. Fixed to 'torsiondrive' since this is the Record explicit to "
        "TorsionDrive.",
    )
    program: _td_constr = Field(
        "torsiondrive",
        description="The name of the program. Fixed to 'torsiondrive' since this is the Record explicit to "
        "TorsionDrive.",
    )

    # Input data
    initial_molecule: List[ObjectId] = Field(..., description="Id(s) of the initial molecule(s) in the database.")
    keywords: TDKeywords = Field(..., description="The TorsionDrive-specific input arguments used for this operation.")
    optimization_spec: OptimizationSpecification = Field(
        ...,
        description="The settings which describe how the energy optimizations at each step of the torsion "
        "scan used for this operation.",
    )
    qc_spec: QCSpecification = Field(
        ...,
        description="The settings which describe how the individual quantum chemistry calculations are handled for "
        "this operation.",
    )

    # Output data
    final_energy_dict: Dict[str, float] = Field(
        ..., description="The final energy at each angle of the TorsionDrive scan."
    )

    optimization_history: Dict[str, List[ObjectId]] = Field(
        ...,
        description="The map of each angle of the TorsionDrive scan to each optimization computations. "
        "Each value of the dict maps to a sequence of :class:`ObjectId` strings which each "
        "point to a single computation in the Database.",
    )
    minimum_positions: Dict[str, int] = Field(  # TODO: This could use review
        ...,
        description="A map of each TorsionDrive angle to the integer index of that angle's optimization "
        "trajectory which has the minimum-energy of the trajectory.",
    )

    class Config(RecordBase.Config):
        pass

    ## Utility

    def _serialize_key(self, key):
        if isinstance(key, str):
            return key
        elif isinstance(key, (int, float)):
            key = (int(key),)

        return json.dumps(key)

    def _deserialize_key(self, key: str) -> Tuple[int, ...]:
        return tuple(json.loads(key))

    def _organize_return(
        self, data: Dict[str, Any], key: Union[int, str, None], minimum: bool = False
    ) -> Dict[str, Any]:

        if key is None:
            return {self._deserialize_key(k): copy.deepcopy(v) for k, v in data.items()}

        key = self._serialize_key(key)

        if minimum:
            minpos = self.minimum_positions[self._serialize_key(key)]
            return copy.deepcopy(data[key][minpos])
        else:
            return copy.deepcopy(data[key])

    def detailed_status(self) -> Dict[str, Any]:

        # Compute the total number of grid points
        tpoints = 1
        for x in self.keywords.grid_spacing:
            tpoints *= int(360 / x)

        flat_history = [x for v in self.get_history().values() for x in v]

        ret = {
            "status": self.status.value,
            "total_points": tpoints,
            "computed_points": len(self.optimization_history),
            "complete_tasks": sum(x.status == "COMPLETE" for x in flat_history),
            "incomplete_tasks": sum((x.status == "INCOMPLETE") or (x.status == "RUNNING") for x in flat_history),
            "error_tasks": sum(x.status == "ERROR" for x in flat_history),
        }
        ret["current_tasks"] = ret["error_tasks"] + ret["incomplete_tasks"]
        ret["percent_complete"] = ret["computed_points"] / ret["total_points"] * 100
        ret["errors"] = [x for x in flat_history if x.status == "ERROR"]

        return ret

    ## Query

    def get_history(
        self, key: Union[int, Tuple[int, ...], str] = None, minimum: bool = False
    ) -> Dict[str, List["ResultRecord"]]:
        """Queries the server for all optimization trajectories.

        Parameters
        ----------
        key : Union[int, Tuple[int, ...], str], optional
            Specifies a single entry to pull from.
        minimum : bool, optional
            If true only returns the minimum optimization, otherwise returns all trajectories.

        Returns
        -------
        Dict[str, List['ResultRecord']]
            The optimization history

        """

        if "history" not in self.cache:

            # Grab procedures
            needed_ids = [x for v in self.optimization_history.values() for x in v]
            objects = []
            for i in range(0, len(needed_ids), self.client.query_limit):
                objects.extend(self.client.query_procedures(id=needed_ids[i : i + self.client.query_limit]))
            procedures = {v.id: v for v in objects}

            # Move procedures into the correct order
            ret = {}
            for okey, hashes in self.optimization_history.items():
                tmp = []
                for h in hashes:
                    tmp.append(procedures[h])
                ret[okey] = tmp

            self.cache["history"] = ret

        data = self.cache["history"]

        return self._organize_return(data, key, minimum=minimum)

    def get_final_energies(self, key: Union[int, Tuple[int, ...], str] = None) -> Dict[str, float]:
        """
        Provides the final optimized energies at each grid point.

        Parameters
        ----------
        key : Union[int, Tuple[int, ...], str], optional
            Specifies a single entry to pull from.

        Returns
        -------
        energy : Dict[str, float]
            Returns energies at each grid point in a dictionary or at a
            single point if a key is specified.

        Examples
        --------
        >>> torsiondrive_obj.get_final_energies()
        {(-90,): -148.7641654446243, (180,): -148.76501336993732, (0,): -148.75056290106735, (90,): -148.7641654446148}

        """

        return self._organize_return(self.final_energy_dict, key)

    def get_final_molecules(self, key: Union[int, Tuple[int, ...], str] = None) -> Dict[str, "Molecule"]:
        """Returns the optimized molecules at each grid point

        Parameters
        ----------
        key : Union[int, Tuple[int, ...], str], optional
            Specifies a single entry to pull from.

        Returns
        -------
        final_molecules : Dict[str, 'Molecule']
            Returns molecule at each grid point in a dictionary or at a
            single point if a key is specified.

        Examples
        --------
        >>> mols = torsiondrive_obj.get_final_molecules()
        >>> type(mols[(-90, )])
        qcelemental.models.molecule.Molecule

        >>> type(torsiondrive_obj.get_final_molecules((-90,)))
        qcelemental.models.molecule.Molecule

        """

        if "final_molecules" not in self.cache:

            map_id_key = self._get_min_optimization_map()

            opt_ids = list(map_id_key.keys())
            results = self.client.custom_query("optimization", "final_molecule", {"optimization_ids": opt_ids})

            ret = {map_id_key[opt_id]: mol_record for opt_id, mol_record in results.items()}

            self.cache["final_molecules"] = ret

        data = self.cache["final_molecules"]

        return self._organize_return(data, key)

    def get_final_results(self, key: Union[int, Tuple[int, ...], str] = None) -> Dict[str, "ResultRecord"]:
        """Returns the final opt gradient result records at each grid point

        Parameters
        ----------
        key : Union[int, Tuple[int, ...], str], optional
            Specifies a single entry to pull from.

        Returns
        -------
        final_results : Dict[str, 'ResultRecord']
            Returns ResultRecord at each grid point in a dictionary or at a
            single point if a key is specified.

        Examples
        --------
        >>> mols = torsiondrive_obj.get_final_results()
        >>> type(mols[(-90, )])
        qcfractal.interface.models.records.ResultRecord

        >>> type(torsiondrive_obj.get_final_results((-90,)))
        qcfractal.interface.models.records.ResultRecord

        """

        if "final_results" not in self.cache:

            map_id_key = self._get_min_optimization_map()
            ret = {}

            # combine the ids into one query
            opt_ids = list(map_id_key.keys())
            results = self.client.custom_query("optimization", "final_result", {"optimization_ids": opt_ids})

            for opt_id, grad_result_record in results.items():
                k = map_id_key[opt_id]
                ret[k] = grad_result_record

            self.cache["final_results"] = ret

        data = self.cache["final_results"]

        return self._organize_return(data, key)

    def visualize(
        self,
        relative: bool = True,
        units: str = "kcal / mol",
        digits: int = 3,
        use_measured_angle: bool = False,
        return_figure: Optional[bool] = None,
    ) -> "plotly.Figure":
        """
        Parameters
        ----------
        relative : bool, optional
            Shows relative energy, lowest energy per scan is zero.
        units : str, optional
            The units of the plot.
        digits : int, optional
            Rounds the energies to n decimal places for display.
        use_measured_angle : bool, optional
            If True, the measured final angle instead of the constrained optimization angle.
            Can provide more accurate results if the optimization was ill-behaved,
            but pulls additional data from the server and may take longer.
        return_figure : Optional[bool], optional
            If True, return the raw plotly figure. If False, returns a hosted iPlot. If None, return a iPlot display in Jupyter notebook and a raw plotly figure in all other circumstances.

        Returns
        -------
        plotly.Figure
            The requested figure.
        """

        # Pull energy dictionary apart
        min_energy = 1e12
        x = []
        y = []
        for k, v in self.get_final_energies().items():
            if len(k) >= 2:
                raise TypeError("TorsionDrive.visualize is currently only available for 1-D scans.")

            if use_measured_angle:
                # Recalculate the dihedral angle
                dihedral_indices = self.keywords.dihedrals[0]
                mol = self.get_final_molecules(k)
                x.append(mol.measure(dihedral_indices))

            else:
                x.append(k[0])

            y.append(v)

            # Update minimum energy
            if v < min_energy:
                min_energy = v

        x = np.array(x)
        y = np.array(y)

        # Sort by angle
        sorter = np.argsort(x)
        x = x[sorter]
        y = y[sorter]
        if relative:
            y -= min_energy

        cf = constants.conversion_factor("hartree", units)
        trace = {"mode": "lines+markers", "x": x, "y": np.around(y * cf, digits)}
        # "name": "something"

        title = "TorsionDrive 1-D Plot"

        if relative:
            ylabel = f"Relative Energy [{units}]"
        else:
            ylabel = f"Absolute Energy [{units}]"

        custom_layout = {
            "title": title,
            "yaxis": {"title": ylabel, "zeroline": True},
            "xaxis": {"title": "Dihedral Angle [degrees]", "zeroline": False, "range": [x.min() - 10, x.max() + 10]},
        }

        return scatter_plot([trace], custom_layout=custom_layout, return_figure=return_figure)

    def _get_min_optimization_map(self):

        map_id_key = {}
        for k, tasks in self.optimization_history.items():
            k = self._serialize_key(k)
            minpos = self.minimum_positions[k]
            final_opt_id = tasks[minpos]
            map_id_key[final_opt_id] = k

        return map_id_key



"""Add collection tables (dataset, reaction_dataset)

Revision ID: 129ff3ce9247
Revises: e32b61e2516f
Create Date: 2019-08-19 16:35:19.100113

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.orm.session import Session

from qcfractal.storage_sockets.models import CollectionORM, DatasetORM, ReactionDatasetORM


# revision identifiers, used by Alembic.
revision = "129ff3ce9247"
down_revision = "e32b61e2516f"
branch_labels = None
depends_on = None


def upgrade():
    print("Start schema migration...")
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "dataset",
        sa.Column("default_benchmark", sa.String(), nullable=True),
        sa.Column("default_keywords", sa.JSON(), nullable=True),
        sa.Column("default_driver", sa.String(), nullable=True),
        sa.Column("default_units", sa.String(), nullable=True),
        sa.Column("alias_keywords", sa.JSON(), nullable=True),
        sa.Column("default_program", sa.String(), nullable=True),
        sa.Column("contributed_values", sa.JSON(), nullable=True),
        sa.Column("history_keys", sa.JSON(), nullable=True),
        sa.Column("history", sa.JSON(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["id"], ["collection.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "reaction_dataset",
        sa.Column("default_benchmark", sa.String(), nullable=True),
        sa.Column("default_keywords", sa.JSON(), nullable=True),
        sa.Column("default_driver", sa.String(), nullable=True),
        sa.Column("default_units", sa.String(), nullable=True),
        sa.Column("alias_keywords", sa.JSON(), nullable=True),
        sa.Column("default_program", sa.String(), nullable=True),
        sa.Column("contributed_values", sa.JSON(), nullable=True),
        sa.Column("history_keys", sa.JSON(), nullable=True),
        sa.Column("history", sa.JSON(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("ds_type", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(["id"], ["collection.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "dataset_entry",
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.Column("molecule_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("comment", sa.String(), nullable=True),
        sa.Column("local_results", sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(["dataset_id"], ["dataset.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(["molecule_id"], ["molecule.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("dataset_id", "molecule_id"),
    )
    op.create_table(
        "reaction_dataset_entry",
        sa.Column("reaction_dataset_id", sa.Integer(), nullable=False),
        sa.Column("attributes", sa.JSON(), nullable=True),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("reaction_results", sa.JSON(), nullable=True),
        sa.Column("stoichiometry", sa.JSON(), nullable=True),
        sa.Column("extras", sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(["reaction_dataset_id"], ["reaction_dataset.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("reaction_dataset_id", "name"),
    )
    op.add_column("collection", sa.Column("collection_type", sa.String(), nullable=True))
    op.add_column("collection", sa.Column("provenance", sa.JSON(), nullable=True))
    op.create_index("ix_collection_type", "collection", ["collection_type"], unique=False)
    ### end Alembic commands ###

    # ------------ copy data
    print("Start Data migration...")
    migrate_collections()


def migrate_collections():
    session = Session(bind=op.get_bind())

    collection_map = {"dataset": DatasetORM, "reactiondataset": ReactionDatasetORM}

    collections = session.query(CollectionORM)

    for collection in collections:
        print(f"collection: id:{collection.id}, lname: {collection.lname}")

        if collection.collection in collection_map.keys():
            collection_class = collection_map[collection.collection]
        else:
            continue

        fields = collection.to_dict(exclude=["id"])
        session.query(CollectionORM).filter_by(id=collection.id).delete(synchronize_session=False)
        session.commit()

        dataset = collection_class(**fields)
        session.add(dataset)
        session.commit()

        dataset.update_relations(**fields)  # will use related attr
        session.commit()

        # print(f'dataset.id: {dataset.id},\n dataset.records: {len(dataset.records)}')


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_collection_type", table_name="collection")
    op.drop_column("collection", "collection_type")
    op.drop_column("collection", "provenance")
    op.drop_table("reaction_dataset_entry")
    op.drop_table("dataset_entry")
    op.drop_table("reaction_dataset")
    op.drop_table("dataset")
    # ### end Alembic commands ###



"""
Tests the server compute capabilities.
"""

import numpy as np
import pytest
import requests
from qcelemental.util import parse_version

import qcengine as qcng
import qcfractal.interface as ptl
from qcfractal import testing
from qcfractal.testing import fractal_compute_server


### Tests the compute queue stack
@testing.using_psi4
def test_compute_queue_stack(fractal_compute_server):

    # Build a client
    client = ptl.FractalClient(fractal_compute_server)

    # Add a hydrogen and helium molecule
    hydrogen = ptl.Molecule.from_data([[1, 0, 0, -0.5], [1, 0, 0, 0.5]], dtype="numpy", units="bohr")
    helium = ptl.Molecule.from_data([[2, 0, 0, 0.0]], dtype="numpy", units="bohr")

    hydrogen_mol_id, helium_mol_id = client.add_molecules([hydrogen, helium])

    kw = ptl.models.KeywordSet(**{"values": {"e_convergence": 1.0e-8}})
    kw_id = client.add_keywords([kw])[0]

    # Add compute
    compute_args = {"driver": "energy", "method": "HF", "basis": "sto-3g", "keywords": kw_id, "program": "psi4"}

    # Ask the server to compute a new computation
    r = client.add_compute("psi4", "HF", "sto-3g", "energy", kw_id, [hydrogen_mol_id, helium])
    assert len(r.ids) == 2

    r2 = client.add_compute(**compute_args, molecule=[hydrogen_mol_id, helium])
    assert len(r2.ids) == 2
    assert len(r2.submitted) == 0
    assert set(r2.ids) == set(r.ids)

    # Manually handle the compute
    fractal_compute_server.await_results()
    assert len(fractal_compute_server.list_current_tasks()) == 0

    # Query result and check against out manual pul
    results_query = {
        "program": "psi4",
        "molecule": [hydrogen_mol_id, helium_mol_id],
        "method": compute_args["method"],
        "basis": compute_args["basis"],
    }
    results = client.query_results(**results_query, status=None)

    assert len(results) == 2
    for r in results:
        assert r.provenance.creator.lower() == "psi4"
        if r.molecule == hydrogen_mol_id:
            assert pytest.approx(-1.0660263371078127, 1e-5) == r.properties.scf_total_energy
        elif r.molecule == helium_mol_id:
            assert pytest.approx(-2.807913354492941, 1e-5) == r.properties.scf_total_energy
        else:
            raise KeyError("Returned unexpected Molecule ID.")

    assert "RHF Reference" in results[0].get_stdout()


### Tests the compute queue stack
@testing.using_psi4
def test_compute_wavefunction(fractal_compute_server):

    psiver = qcng.get_program("psi4").get_version()
    if parse_version(psiver) < parse_version("1.4a2.dev160"):
        pytest.skip("Must be used a modern version of Psi4 to execute")

    # Build a client
    client = ptl.FractalClient(fractal_compute_server)

    # Add a hydrogen and helium molecule
    hydrogen = ptl.Molecule.from_data([[1, 0, 0, -0.5], [1, 0, 0, 0.5]], dtype="numpy", units="bohr")

    # Ask the server to compute a new computation
    r = client.add_compute(
        program="psi4",
        driver="energy",
        method="HF",
        basis="sto-3g",
        molecule=hydrogen,
        protocols={"wavefunction": "orbitals_and_eigenvalues"},
    )

    fractal_compute_server.await_results()
    assert len(fractal_compute_server.list_current_tasks()) == 0

    result = client.query_results(id=r.ids)[0]
    assert result.wavefunction

    r = result.get_wavefunction("orbitals_a")
    assert isinstance(r, np.ndarray)
    assert r.shape == (2, 2)

    r = result.get_wavefunction(["orbitals_a", "basis"])
    assert r.keys() == {"orbitals_a", "basis"}


### Tests the compute queue stack
@testing.using_geometric
@testing.using_psi4
def test_procedure_optimization_single(fractal_compute_server):

    # Add a hydrogen molecule
    hydrogen = ptl.Molecule.from_data([[1, 0, 0, -0.672], [1, 0, 0, 0.672]], dtype="numpy", units="bohr")
    client = ptl.FractalClient(fractal_compute_server.get_address(""))
    mol_ret = client.add_molecules([hydrogen])

    kw = ptl.models.KeywordSet(values={"scf_properties": ["quadrupole", "wiberg_lowdin_indices"]})
    kw_id = client.add_keywords([kw])[0]

    # Add compute
    options = {
        "keywords": None,
        "qc_spec": {"driver": "gradient", "method": "HF", "basis": "sto-3g", "keywords": kw_id, "program": "psi4"},
    }

    # Ask the server to compute a new computation
    r = client.add_procedure("optimization", "geometric", options, mol_ret)
    assert len(r.ids) == 1
    compute_key = r.ids[0]

    # Manually handle the compute
    fractal_compute_server.await_results()
    assert len(fractal_compute_server.list_current_tasks()) == 0

    # # Query result and check against out manual pul
    query1 = client.query_procedures(procedure="optimization", program="geometric")
    query2 = client.query_procedures(id=compute_key)

    for query in [query1, query2]:
        assert len(query) == 1
        opt_result = query[0]

        assert isinstance(opt_result.provenance.creator, str)
        assert isinstance(str(opt_result), str)  # Check that repr runs
        assert pytest.approx(-1.117530188962681, 1e-5) == opt_result.get_final_energy()

        # Check pulls
        traj = opt_result.get_trajectory()
        assert len(traj) == len(opt_result.energies)

        assert np.array_equal(opt_result.get_final_molecule().symbols, ["H", "H"])

        # Check individual elements
        for ind in range(len(opt_result.trajectory)):
            assert traj[ind].program == "psi4"

            # Check keywords went through
            assert traj[ind].provenance.creator.lower() == "psi4"
            assert int(traj[ind].keywords) == int(kw_id)
            assert "SCF QUADRUPOLE XY" in traj[ind].extras["qcvars"]
            assert "WIBERG_LOWDIN_INDICES" in traj[ind].extras["qcvars"]

            # Make sure extra was popped
            assert "_qcfractal_tags" not in traj[ind].extras

            raw_energy = traj[ind].properties.return_energy
            assert pytest.approx(raw_energy, 1.0e-5) == opt_result.energies[ind]

        # Check result stdout
        assert "RHF Reference" in traj[0].get_stdout()

        assert opt_result.get_molecular_trajectory()[0].id == opt_result.initial_molecule
        assert opt_result.get_molecular_trajectory()[-1].id == opt_result.final_molecule

        # Check stdout
        assert "internal coordinates" in opt_result.get_stdout()

    # Check that duplicates are caught
    r = client.add_procedure("optimization", "geometric", options, [mol_ret[0]])
    assert len(r.ids) == 1
    assert len(r.existing) == 1


@testing.using_geometric
@testing.using_psi4
def test_procedure_optimization_protocols(fractal_compute_server):

    # Add a hydrogen molecule
    hydrogen = ptl.Molecule.from_data([[1, 0, 0, -0.673], [1, 0, 0, 0.673]], dtype="numpy", units="bohr")
    client = fractal_compute_server.client()

    # Add compute
    options = {
        "keywords": None,
        "qc_spec": {"driver": "gradient", "method": "HF", "basis": "sto-3g", "program": "psi4"},
        "protocols": {"trajectory": "final"},
    }

    # Ask the server to compute a new computation
    r = client.add_procedure("optimization", "geometric", options, [hydrogen])
    assert len(r.ids) == 1
    compute_key = r.ids[0]

    # Manually handle the compute
    fractal_compute_server.await_results()
    assert len(fractal_compute_server.list_current_tasks()) == 0

    # # Query result and check against out manual pul
    proc = client.query_procedures(id=r.ids)[0]
    assert proc.status == "COMPLETE"

    assert len(proc.trajectory) == 1
    assert len(proc.energies) > 1



"""Make some required fields not nullable

Revision ID: 5be555fe9dc0
Revises: 5686f64a46c8
Create Date: 2020-10-29 15:12:53.172977

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB


# revision identifiers, used by Alembic.
revision = "5be555fe9dc0"
down_revision = "5686f64a46c8"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("grid_optimization_procedure", "initial_molecule", existing_type=sa.INTEGER(), nullable=False)
    op.alter_column("optimization_procedure", "initial_molecule", existing_type=sa.INTEGER(), nullable=False)
    op.alter_column("base_result", "procedure", existing_type=sa.VARCHAR(length=100), nullable=False)
    op.alter_column("base_result", "protocols", existing_type=JSONB(astext_type=sa.Text()), nullable=False)
    op.alter_column("result", "molecule", existing_type=sa.INTEGER(), nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("base_result", "protocols", existing_type=JSONB(astext_type=sa.Text()), nullable=True)
    op.alter_column("base_result", "procedure", existing_type=sa.VARCHAR(length=100), nullable=True)
    op.alter_column("result", "molecule", existing_type=sa.INTEGER(), nullable=True)
    op.alter_column("optimization_procedure", "initial_molecule", existing_type=sa.INTEGER(), nullable=True)
    op.alter_column("grid_optimization_procedure", "initial_molecule", existing_type=sa.INTEGER(), nullable=True)
    # ### end Alembic commands ###
